from __future__ import annotations
import dataclasses
import copy
import inspect
import logging
import numpy as np
from typing import Callable, Dict, List, Optional, Tuple, TYPE_CHECKING
from .utils import Array, ArrayLike


if TYPE_CHECKING:
    num_samples = float
    num_values = float
    num_spectrum_pts = float
    num_test_pts = float
    num_y_channels = float


logger = logging.getLogger(__name__)


@dataclasses.dataclass
class ModelParameter:
    """Metadata associated with a model parameter.

    Attributes:
        lower_bound: lower bound for the parameter. Fitted values are guaranteed to be
            greater than or equal to the lower bound. Parameter bounds may be used by
            fit heuristics to help find good starting points for the optimizer.
        upper_bound: upper bound for the parameter. Fitted values are guaranteed to be
            lower than or equal to the upper bound. Parameter bounds may be used by
            fit heuristics to help find good starting points for the optimizer.
        fixed_to: if not `None`, the model parameter is fixed to this value during
            fitting instead of being floated. This value may additionally be used by
            the heuristics to help find good initial values for other model parameters
            for which none have been provided by the user. The value of `fixed_to` must
            lie within the bounds of the parameter.
        user_estimate: if not `None` and the parameter is not fixed, this value is
            used as an initial value during fitting rather than obtaining a value from
            the heuristics. This value may additionally be used by the heuristics to
            help find good initial values for other model parameters for which none
            have been provided by the user. The value of `user_estimate` must lie
            within the bounds of the parameter.
        heuristic: if both of `fixed_to` and `user_estimate` are `None`, this value is
            used as an initial value during fitting. It is set by the
            `estimate_parameters` method of the model in which the parameter is used
            and should not be set by the user.
        scale_func: callable returning a scale factor which the parameter must be
            *multiplied* by if it was fitted using `x` / `y` data that has been
            *multiplied* by the given scale factors. Scale factors are used to improve
            numerical stability by avoiding asking the optimizer to work with very large
            or very small values of `x` and `y`. The callable takes three arguments: the
            x-axis and y-axis scale factors and the model instance. If any `scale_func`
            returns a value that is `0`, not finite (e.g. `nan`) or `None` we do not
            rescale the coordinates. c.f. :meth can_rescale: and :meth rescale:.
    """

    lower_bound: float = -np.inf
    upper_bound: float = np.inf
    fixed_to: Optional[float] = None
    user_estimate: Optional[float] = None
    heuristic: Optional[float] = None
    scale_factor: float = 1
    scale_func: Callable[
        [
            float,
            float,
            Model,
        ],
        Optional[float],
    ] = lambda x_scale, y_scale, model: 1

    def can_rescale(self, x_scale: float, y_scale: float, model: Model) -> bool:
        """Returns `True` if the parameter can be rescaled."""
        scale_factor = self.scale_func(x_scale, y_scale, model)
        return (
            scale_factor is not None and np.isfinite(scale_factor) and scale_factor != 0
        )

    def rescale(self, x_scale: float, y_scale: float, model: Model):
        """Rescales the parameter metadata based on the specified x and y data scale
        factors and returns the overall scale factor used.
        """
        scale_factor = self.scale_func(x_scale, y_scale, model)

        if scale_factor is None:
            raise ValueError("Scale factor must not be None during rescale")

        def _rescale(attr):
            if attr is None:
                return None
            return attr / scale_factor

        self.lower_bound = _rescale(self.lower_bound)
        self.upper_bound = _rescale(self.upper_bound)
        self.fixed_to = _rescale(self.fixed_to)
        self.user_estimate = _rescale(self.user_estimate)
        self.scale_factor *= scale_factor

    def get_initial_value(self) -> float:
        """
        Get initial value.

        For fixed parameters, this is the value the parameter is fixed to. For floated
        parameters, it is the value used to seed the fit. In the latter case, the
        initial value is retrieved from `user_estimate` if that attribute is not
        `None`, otherwise `heuristic` is used.
        """
        if self.fixed_to is not None:
            value = self.fixed_to
        elif self.user_estimate is not None:
            value = self.user_estimate
        elif self.heuristic is not None:
            value = self.clip(self.heuristic)
        else:
            raise ValueError("No initial value specified")

        if value < self.lower_bound or value > self.upper_bound:
            raise ValueError("Initial value outside bounds.")

        return value

    def has_user_initial_value(self) -> bool:
        """Returns True if the parameter is fixed or has a user estimate"""
        return self.fixed_to is not None or self.user_estimate is not None

    def clip(self, value: float) -> float:
        """Clip value to lie between lower and upper bounds."""
        return np.clip(value, self.lower_bound, self.upper_bound)


class Model:
    """Base class for fit models.

    A model groups a function to be fitted with associated metadata (parameter names,
    default bounds etc) and heuristics. It is agnostic about the method of fitting or
    the data statistics.
    """

    def __init__(self, parameters: Optional[Dict[str, ModelParameter]] = None):
        """
        :param parameters: optional dictionary mapping names of model parameters to
            their metadata. This should be `None` (default) if the model has a static
            set of parameters in which case the parameter dictionary is generated from
            the call signature of :meth _func:. The model parameters are stored as
            `self.parameters` and may be modified after construction to change the model
            behaviour during fitting (e.g. to change the bounds, fixed parameters, etc).
        """
        if parameters is None:
            spec = inspect.getfullargspec(self._func)
            for name in spec.args[2:]:
                assert isinstance(
                    spec.annotations[name], ModelParameter
                ), "Model parameters must be instances of `ModelParameter`"
            self.parameters = {
                name: copy.deepcopy(spec.annotations[name]) for name in spec.args[2:]
            }
        else:
            self.parameters = parameters

    def can_rescale(self, x_scale: float, y_scale: float) -> bool:
        """Returns True if the model can be rescaled"""
        return all(
            [
                param_data.can_rescale(x_scale, y_scale, self)
                for param_data in self.parameters.values()
            ]
        )

    @staticmethod
    def get_scaled_model(model, x_scale: float, y_scale: float):
        """Returns a scaled copy of a given model object

        :param model: model to be copied and rescaled
        :param x_scale: x-axis scale factor
        :param y_scale: y-axis scale factor
        :returns: a scaled copy of model
        """
        scaled_model = copy.deepcopy(model)
        for param in scaled_model.parameters.values():
            param.rescale(x_scale, y_scale, scaled_model)

        return scaled_model

    def get_num_y_channels(self) -> int:
        """Returns the number of y channels supported by the model."""
        raise NotImplementedError

    def func(
        self, x: Array[("num_samples",), np.float64], param_values: Dict[str, float]
    ) -> Array[("num_y_channels", "num_samples"), np.float64]:
        """Evaluates the model at a given set of x-axis points and with a given set of
        parameter values and returns the result.

        Overload this to provide a model function with a dynamic set of parameters,
        otherwise prefer to override `_func`.

        :param x: x-axis data
        :param param_values: dictionary of parameter values
        :returns: array of model values
        """
        return self._func(x, **param_values)

    def _func(
        self,
        x: Array[("num_samples",), np.float64],
    ) -> Array[("num_y_channels", "num_samples"), np.float64]:
        """Evaluates the model at a given set of x-axis points and with a given set of
        parameter values and returns the result.

        Overload this in preference to `func` unless the FitModel takes a
        dynamic set of parameters. Use : class ModelParameter: objects as the annotation
        for the parameters arguments. e.g.:

        ```
        def _func(self, x, a: ModelParameter(), b: ModelParameter()):
        ```

        A dictionary of `ModelParameter`s may be accessed via the `self.parameters`
        attribute. These parameters may be modified by the user to change the model
        behaviour during fitting (e.g. to change the bounds, fixed parameters, etc).

        :param x: x-axis data
        :returns: array of model values
        """
        raise NotImplementedError

    def estimate_parameters(
        self,
        x: Array[("num_samples",), np.float64],
        y: Array[("num_y_channels", "num_samples"), np.float64],
        model_parameters: Dict[str, ModelParameter],
    ):
        """Set heuristic values for model parameters.

        Typically called during `Fitter.fit`. This method may make use of information
        supplied by the user for some parameters (via the `fixed_to` or
        `user_estimate` attributes) to find initial guesses for other parameters.

        The datasets must be sorted in order of increasing x-axis values and must not
        contain any infinite or nan values. If all parameters of the model allow
        rescaling, then `x`, `y` and `model_parameters` will contain rescaled values.

        TODO: this should act directly on self.model_parameters rather than taking
        model parameters as an argument (this is a hangover from an old design)

        :param x: x-axis data, rescaled if allowed.
        :param y: y-axis data, rescaled if allowed.
        :param model_parameters: dictionary mapping model parameter names to their
            metadata, rescaled if allowed.
        """
        raise NotImplementedError

    def calculate_derived_params(
        self,
        x: Array[("num_samples",), np.float64],
        y: Array[("num_y_channels", "num_samples"), np.float64],
        fitted_params: Dict[str, float],
        fit_uncertainties: Dict[str, float],
    ) -> Tuple[Dict[str, float], Dict[str, float]]:
        """
        Returns dictionaries of values and uncertainties for the derived model
        parameters (parameters which are calculated from the fit results rather than
        being directly part of the fit) based on values of the fitted parameters and
        their uncertainties.

        :param x: x-axis data
        :param y: y-axis data
        :param: fitted_params: dictionary mapping model parameter names to their
            fitted values.
        :param fit_uncertainties: dictionary mapping model parameter names to
            their fit uncertainties.
        :returns: tuple of dictionaries mapping derived parameter names to their
            values and uncertainties.
        """
        return {}, {}

    def param_min_sqrs(
        self,
        x: Array[("num_samples",), np.float64],
        y: Array[("num_y_channels", "num_samples"), np.float64],
        parameters: Dict[str, ModelParameter],
        scanned_param: str,
        scanned_param_values: ArrayLike["num_values", np.float64],
    ) -> Tuple[float, float]:
        """Scans one model parameter while holding the others fixed to find the value
        that gives the best fit to the data (minimum sum-squared residuals).

        :param x: x-axis data
        :param y: y-axis data
        :param parameters: dictionary of model parameters. All parameters apart from the
          scanned parameter must have an initial value set.
        :param scanned_param: name of parameter to optimize
        :param scanned_param_values: array of scanned parameter values to test

        :returns: tuple with the value from :param scanned_param_values: which results
        in lowest residuals and the root-sum-squared residuals for that value.
        """
        param_values = {
            param: param_data.get_initial_value()
            for param, param_data in parameters.items()
            if param != scanned_param
        }

        scanned_param_values = np.asarray(scanned_param_values)
        costs = np.zeros(scanned_param_values.shape)
        for idx, value in np.ndenumerate(scanned_param_values):
            param_values[scanned_param] = value
            y_params = self.func(x, param_values)
            costs[idx] = np.sqrt(np.sum(np.power(y - y_params, 2)))
        opt = np.argmin(costs)
        return float(scanned_param_values[opt]), float(costs[opt])

    def find_x_offset_sym_peak(
        self,
        x: Array[("num_samples",), np.float64],
        y: Array[("num_samples",), np.float64],
        parameters: Dict[str, ModelParameter],
        omega: Array[("num_spectrum_pts",), np.float64],
        spectrum: Array[("num_spectrum_pts",), np.float64],
        omega_cut_off: float,
        test_pts: Optional[Array[("num_test_pts",), np.float64]] = None,
        x_offset_param_name: str = "x0",
        y_offset_param_name: str = "y0",
    ):
        """Finds the x-axis offset for symmetric, peaked (maximum deviation from the
        baseline occurs at the origin) functions.

        This heuristic draws candidate x-offset points from three sources and picks the
        best one (in the least-squares residuals sense). Sources:
          - FFT shift theorem based on provided spectrum data
          - Tests all points in the top quartile of deviation from the baseline
          - Optionally, user-provided "test points", taken from another heuristic. This
            allows the developer to combine the general-purpose heuristics here with
            other heuristics which make use of more model-specific assumptions

        :param x: x-axis data
        :param y: y-axis data. For models with multiple y channels, this should contain
            data from a single channel only.
        :param parameters: dictionary of model parameters. All model parameters other
          than the x-axis offset must have initial values set before calling this method
        :param omega: FFT frequency axis
        :param spectrum: complex FFT data. For models with multiple y channels, this
          should contain data from a single channel only.
        :param omega_cut_off: highest value of omega to use in offset estimation
        :param test_pts: optional array of x-axis points to test
        :param x_offset_param_name: name of the x-axis offset model parameter
        :param y_offset_param_name: name of the y-axis offset model parameter

        :returns: an estimate of the x-axis offset
        """
        if y.ndim != 1:
            raise ValueError(
                f"{y.shape[0]} y-channels were provided to a method which takes 1."
            )

        x0_candidates = np.array([])

        if test_pts is not None:
            x0_candidates = np.append(x0_candidates, test_pts)

        try:
            fft_candidate = self.find_x_offset_fft(
                x=x, omega=omega, spectrum=spectrum, omega_cut_off=omega_cut_off
            )
            x0_candidates = np.append(x0_candidates, fft_candidate)
        except ValueError:
            pass

        y0 = parameters[y_offset_param_name].get_initial_value()
        deviations = np.argsort(np.abs(y - y0))
        top_quartile_deviations = deviations[int(len(deviations) * 3 / 4) :]
        deviations_candidates = x[top_quartile_deviations]
        x0_candidates = np.append(x0_candidates, deviations_candidates)

        best_x0, _ = self.param_min_sqrs(
            x=x,
            y=y,
            parameters=parameters,
            scanned_param=x_offset_param_name,
            scanned_param_values=x0_candidates,
        )
        return best_x0

    def find_x_offset_sampling(
        self,
        x: Array[("num_samples",), np.float64],
        y: Array[("num_samples", "num_y_channels"), np.float64],
        parameters: Dict[str, ModelParameter],
        width: float,
        x_offset_param_name: str = "x0",
    ) -> float:
        """Finds the x-axis offset of a dataset by stepping through a range of potential
        offset values and picking the one that gives the lowest residuals.

        This method is typically called during parameter estimation after all other
        model parameters have been estimated from the periodogram (which is not itself
        sensitive to offsets).

        There are a few ways one can implemented this functionality: for strongly
        peaked functions we could have used a simple peak search; we could have used an
        FFT, fitted a line to the phase and used the Fourier Transform Shift Theorem.

        This function takes a more brute-force approach by evaluating the model at a
        range of offset values, picking the one that gives the lowest residuals. This
        may be appropriate where one needs the estimate to be highly robust in the face
        of noisy, irregularly sampled data.

        :param x: x-axis data
        :param y: y-axis data
        :param parameters: dictionary of model parameters.
        :param width: width of the feature we're trying to find (e.g. FWHMH). Used to
            pick the spacing between offset values to try.
        :param x_offset_param_name: name of the x-axis offset parameter

        :returns: an estimate of the x-axis offset
        """
        offsets = np.arange(min(x), max(x), width / 6)
        return self.param_min_sqrs(x, y, parameters, x_offset_param_name, offsets)[0]

    def find_x_offset_fft(
        self,
        x: Array[("num_samples",), np.float64],
        omega: Array[("num_spectrum_pts",), np.float64],
        spectrum: Array[("num_spectrum_pts",), np.float64],
        omega_cut_off: float,
    ) -> float:
        """Finds the x-axis offset of a dataset from the phase of an FFT.

        This method is typically called during parameter estimation after all other
        model parameters have been estimated from the periodogram (which is not itself
        sensitive to offsets).

        This method uses the FFT shift theorem to extract the offset from the phase
        slope of an FFT. At present it only supports models with a single y channel.

        :param omega: FFT frequency axis
        :param spectrum: complex FFT data. For models with multiple y channels, this
          should contain data from a single channel only.
        :param omega_cut_off: highest value of omega to use in offset estimation

        :returns: an estimate of the x-axis offset
        """
        if spectrum.ndim != 1:
            raise ValueError(
                f"{spectrum.shape[1]} y channels were provided to a method which "
                "takes 1"
            )

        keep = omega < omega_cut_off
        if np.sum(keep) < 2:
            raise ValueError("Insufficient data below cut-off")

        omega = omega[keep]
        phi = np.unwrap(np.angle(spectrum[keep]))
        phi -= phi[0]

        p = np.polyfit(omega, phi, deg=1)

        x0 = min(x) - p[0]
        x0 = x0 if x0 > min(x) else x0 + x.ptp()
        return x0

    def get_initial_values(
        self, model_parameters: Optional[Dict[str, ModelParameter]] = None
    ) -> Dict[str, float]:
        """Returns a dictionary mapping model parameter names to their initial values.

        :param model_parameters: optional dictionary mapping model parameter names to
           :class ModelParameter:s
        """
        model_paramers = model_parameters or self.parameters
        return {
            param: param_data.get_initial_value()
            for param, param_data in model_paramers.items()
        }


class Fitter:
    """Base class for fitters.

    Fitters perform maximum likelihood parameter estimation on a dataset under the
    assumption of a certain model and statistics (normal, binomial, etc) and store the
    results as attributes.

    Attributes:
        x: 1D ndarray of shape (num_samples,) containing x-axis values of valid points.
        y: 2D ndarray of shape (num_y_channels, num_samples) containing y-axis values
            of valid points.
        sigma: optional 2D ndarray of shape (num_y_channels, num_samples) containing
            y-axis standard deviations of valid points. Only used by `NormalFitter`.
        values: dictionary mapping model parameter names to their fitted values
        uncertainties: dictionary mapping model parameter names to their fit
            uncertainties. For sufficiently large datasets, well-formed problems and
            ignoring covariances these are the 1-sigma confidence intervals (roughly:
            there is a 1/3 chance that the real parameter values differ from their
            fitted values by more than this much)
        derived_values: dictionary mapping names of derived parameters (parameters which
            are not part of the fit, but are calculated by the model from the fitted
            parameter values) to their values
        derived_uncertainties: dictionary mapping names of derived parameters to their
            fit uncertainties
        initial_values: dictionary mapping model parameter names to the initial values
            used to seed the fit.
        model: the fit model
        fit_significance: if applicable, the fit significance as a number between 0 and
            1 (1 indicates perfect agreement between the fitted model and input
            dataset). See :meth _fit_significance: for details.
        free_parameters: list of names of the model parameters floated during the fit
    """

    x: Array[("num_samples",), np.float64]
    y: Array[("num_y_channels", "num_samples"), np.float64]
    sigma: Optional[Array[("num_y_channels", "num_samples"), np.float64]]
    values: Dict[str, float]
    uncertainties: Dict[str, float]
    derived_values: Dict[str, float]
    derived_uncertainties: Dict[str, float]
    initial_values: Dict[str, float]
    model: Model
    fit_significance: Optional[float]
    free_parameters: List[str]

    def __init__(
        self,
        x: ArrayLike[("num_samples",), np.float64],
        y: ArrayLike[("num_y_channels", "num_samples"), np.float64],
        model: Model,
        sigma: Optional[
            ArrayLike[("num_y_channels", "num_samples"), np.float64]
        ] = None,
    ):
        """Fits a model to a dataset and stores the results.

        :param x: x-axis data
        :param y: y-axis data
        :param sigma: optional y-axis  standard deviations. Only used by `NormalFitter`
        :param model: the model function to fit to. The model's parameter dictionary is
            used to configure the fit (set parameter bounds etc). Modify this before
            fitting to change the fit behaviour from the model class' defaults.
        """
        model = copy.deepcopy(model)

        # Sanitize input dataset
        x = np.array(x, dtype=np.float64, copy=True)
        y = np.array(y, dtype=np.float64, copy=True)
        sigma = None if sigma is None else np.array(sigma, dtype=np.float64, copy=True)

        if x.ndim != 1:
            raise ValueError("x-axis data must be a 1D array.")

        if y.ndim > 2:
            raise ValueError("y-axis data must be a 1D or 2D array.")

        # Ensure a common shape for all data related to y-axis, regardless of
        # user input
        y = np.atleast_2d(y)
        if sigma is not None:
            sigma = np.atleast_2d(sigma)

        if x.shape[0] != y.shape[1]:
            raise ValueError(
                "Number of x-axis and y-axis samples must match "
                f"(got {x.shape} and {y.shape})."
            )
        if y.shape[0] != model.get_num_y_channels():
            raise ValueError(
                f"Expected {model.get_num_y_channels()} y-channels, got {y.shape[1]}."
            )

        if sigma is not None and sigma.shape != y.shape:
            raise ValueError(
                f"Shapes of sigma (got {sigma.shape}) and y (got {y.shape}) must match."
            )

        if sigma is not None and not np.all(sigma != 0):
            logger.warning("Ignoring points with zero uncertainty.")

        valid_x = np.isfinite(x)
        valid_y = np.isfinite(y)
        valid_sigma = (
            None if sigma is None else np.logical_and(np.isfinite(sigma), sigma != 0)
        )

        valid_y = np.all(valid_y, axis=0)
        valid_sigma = None if sigma is None else np.all(valid_sigma, axis=0)

        valid_pts = np.logical_and(valid_x, valid_y)
        if sigma is not None:
            valid_pts = np.logical_and(valid_pts, valid_sigma)
        valid_pts = valid_pts.squeeze()

        x = x[valid_pts]
        y = y[:, valid_pts]
        sigma = None if sigma is None else sigma[:, valid_pts]

        inds = np.argsort(x)
        x = x[inds]
        y = y[:, inds]
        sigma = None if sigma is None else sigma[:, inds]

        self.x = x
        self.y = y
        self.sigma = sigma

        # Rescale coordinates to improve numerics (optimizers need to do things like
        # calculate numerical derivatives which is easiest if x and y are O(1)).
        #
        # Currently we use a single scale factor for all y channels. This may change in
        # future
        x_scale = np.max(np.abs(x))
        y_scale = np.max(np.abs(y))

        if model.can_rescale(x_scale, y_scale):
            scaled_model = model.get_scaled_model(model, x_scale, y_scale)
        else:
            x_scale = 1
            y_scale = 1
            scaled_model = copy.deepcopy(model)

        x = x / x_scale
        y = y / y_scale
        sigma = None if sigma is None else sigma / y_scale

        scaled_model.estimate_parameters(x, y, scaled_model.parameters)

        for param, param_data in scaled_model.parameters.items():
            try:
                param_data.get_initial_value()
            except ValueError:
                raise RuntimeError(
                    "No fixed_to, user_estimate or heuristic specified"
                    f" for parameter `{param}`."
                )

        self.fixed_parameters = {
            param_name: param_data.fixed_to
            for param_name, param_data in scaled_model.parameters.items()
            if param_data.fixed_to is not None
        }
        self.free_parameters = [
            param_name
            for param_name, param_data in scaled_model.parameters.items()
            if param_data.fixed_to is None
        ]

        if self.free_parameters == []:
            raise ValueError("Attempt to fit with no free parameters.")

        def free_func(
            x: Array[("num_samples",), np.float64], *free_param_values: float
        ) -> Array[("num_y_channels", "num_samples"), np.float64]:
            """Call the model function with the values of the free parameters."""
            params = {
                param: value
                for param, value in zip(self.free_parameters, list(free_param_values))
            }
            params.update(self.fixed_parameters)
            y = scaled_model.func(x, params)
            return y

        fitted_params, uncertainties = self._fit(
            x, y, sigma, scaled_model.parameters, free_func
        )
        fitted_params.update(
            {param: value for param, value in self.fixed_parameters.items()}
        )
        uncertainties.update({param: 0 for param in self.fixed_parameters.keys()})

        fitted_params = {
            param: value * scaled_model.parameters[param].scale_factor
            for param, value in fitted_params.items()
        }
        uncertainties = {
            param: value * scaled_model.parameters[param].scale_factor
            for param, value in uncertainties.items()
        }

        initial_values = {
            param: param_data.get_initial_value()
            * scaled_model.parameters[param].scale_factor
            for param, param_data in scaled_model.parameters.items()
        }

        (
            self.derived_values,
            self.derived_uncertainties,
        ) = model.calculate_derived_params(self.x, self.y, fitted_params, uncertainties)

        self.model = model
        self.values = fitted_params
        self.uncertainties = uncertainties
        self.initial_values = initial_values
        self.fit_significance = self._fit_significance()

    @staticmethod
    def _fit(
        x: Array[("num_samples",), np.float64],
        y: Array[("num_y_channels", "num_samples"), np.float64],
        sigma: Optional[Array[("num_y_channels", "num_samples"), np.float64]],
        parameters: Dict[str, ModelParameter],
        free_func: Callable[..., Array[("num_y_channels", "num_samples"), np.float64]],
    ) -> Tuple[Dict[str, float], Dict[str, float]]:
        """Implementation of the parameter estimation.

        `Fitter` implementations must override this method to provide a fit with
        appropriate statistics.

        :param x: rescaled x-axis data, must be a 1D array
        :param y: rescaled y-axis data
        :param sigma: rescaled standard deviations
        :param parameters: dictionary of rescaled model parameters
        :param free_func: convenience wrapper for the model function, taking only values
            for the fit's free parameters

        :returns: tuple of dictionaries mapping model parameter names to their fitted
            values and uncertainties.
        """
        raise NotImplementedError

    def _fit_significance(self) -> Optional[float]:
        """Returns an estimate of the goodness of fit as a number between 0 and 1.

        This is the defined as the probability that fit residuals as large as the ones
        we observe could have arisen through chance given our assumed statistics and
        assuming that the fitted model perfectly represents the probability distribution

        A value of `1` indicates a perfect fit (all data points lie on the fitted curve)
        a value close to 0 indicates significant deviations of the dataset from the
        fitted model.
        """
        return None

    def evaluate(
        self,
        transpose_and_squeeze=False,
        x_fit: Optional[Array[("num_samples",), np.float64]] = None,
    ) -> Tuple[
        Array[("num_samples",), np.float64],
        Array[("num_y_channels", "num_samples"), np.float64],
    ]:
        """Evaluates the model function using the fitted parameter set.

        :param transpose_and_squeeze: if True, array `y_fit` is transposed
            and squeezed before being returned. This is intended to be used
            for plotting, since matplotlib requires different y-series to be
            stored as columns.
        :param x_fit: optional x-axis points to evaluate the model at. If
            `None`, we use the values stored as attribute `x` of the fitter.

        :returns: tuple of x-axis values used and corresponding y-axis values
            of the fitted model
        """
        x_fit = x_fit if x_fit is not None else self.x
        y_fit = self.model.func(x_fit, self.values)

        if transpose_and_squeeze:
            return x_fit, y_fit.T.squeeze()
        return x_fit, y_fit

    def residuals(self) -> Array[("num_y_channels", "num_samples"), np.float64]:
        """Returns an array of fit residuals."""
        return self.y - self.evaluate()[1]



import logging
import numpy as np
import pprint
from scipy import optimize, stats
from typing import Callable, Dict, Optional, Tuple, TYPE_CHECKING

from . import Fitter, ModelParameter
from .utils import Array

if TYPE_CHECKING:
    num_samples = float
    num_values = float
    num_y_channels = float
    num_samples_flattened = float


logger = logging.getLogger(__name__)


class NormalFitter(Fitter):
    """Fitter for normally-distributed data.

    We use least-squares fitting as a maximum-likelihood parameter estimator for
    normally distributed data. For data that is close to normal this is usually a pretty
    good approximation of a true MLE estimator. YMMV...
    """

    @staticmethod
    def _fit(
        x: Array[("num_samples",), np.float64],
        y: Array[("num_y_channels", "num_samples"), np.float64],
        sigma: Optional[Array[("num_y_channels", "num_samples"), np.float64]],
        parameters: Dict[str, ModelParameter],
        free_func: Callable[..., Array[("num_y_channels", "num_samples"), np.float64]],
    ) -> Tuple[Dict[str, float], Dict[str, float]]:
        """Implementation of the parameter estimation.

        `Fitter` implementations must override this method to provide a fit with
        appropriate statistics.

        :param x: rescaled x-axis data, must be a 1D array
        :param y: rescaled y-axis data
        :param sigma: rescaled standard deviations
        :param parameters: dictionary of rescaled model parameters
        :param free_func: convenience wrapper for the model function, taking only values
            for the fit's free parameters

        :returns: tuple of dictionaries mapping model parameter names to their fitted
            values and uncertainties.
        """
        free_parameters = [
            param_name
            for param_name, param_data in parameters.items()
            if param_data.fixed_to is None
        ]

        p0 = [parameters[param].get_initial_value() for param in free_parameters]
        lower = [parameters[param].lower_bound for param in free_parameters]
        upper = [parameters[param].upper_bound for param in free_parameters]
        p0_dict = {
            param: parameters[param].get_initial_value() for param in free_parameters
        }

        assert x.dtype == np.float64
        assert y.dtype == np.float64

        logger.debug(
            "Starting least-squares fitting with initial parameters: "
            f"{pprint.pformat(p0_dict)}"
        )

        def fit_func(
            _: Array[("num_samples_flattened",), np.float64], *free_param_values: float
        ) -> Array[("num_samples_flattened",), np.float64]:
            """Call the model function with the values of the free parameters."""
            return np.ravel(free_func(x, *free_param_values))

        x_flat = np.ravel(np.tile(x, (y.ndim, 1)))
        y_flat = np.ravel(y)
        sigma_flat = None if sigma is None else np.ravel(sigma)

        p, p_cov = optimize.curve_fit(
            f=fit_func,
            xdata=x_flat,
            ydata=y_flat,
            p0=p0,
            sigma=sigma_flat,
            absolute_sigma=sigma is not None,
            bounds=(lower, upper),
            method="trf",
        )

        p_err = np.sqrt(np.diag(p_cov))

        p = {param: value for param, value in zip(free_parameters, p)}
        p_err = {param: value for param, value in zip(free_parameters, p_err)}

        return p, p_err

    def chi_squared(
        self,
        x: Array[("num_samples",), np.float64],
        y: Array[("num_samples", "num_y_channels"), np.float64],
        sigma: Array[("num_samples", "num_y_channels"), np.float64],
    ) -> float:
        """Returns the Chi-squared fit significance for the fit compared to a given
        dataset as a number between 0 and 1.

        The significance gives the probability that fit residuals as large as the ones
        we observe could have arisen through chance given our assumed statistics and
        assuming that the fitted model perfectly represents the probability
        distribution.

        A value of `1` indicates a perfect fit (all data points lie on the fitted curve)
        a value close to 0 indicates super-statistical deviations of the dataset from
        the fitted model.
        """
        if sigma.shape != y.shape:
            raise ValueError(
                f"Mismatch between shapes of sigma ({sigma.shape}) and y ({y.shape})"
            )

        n = y.size - len(self.free_parameters)

        if n < 1:
            raise ValueError(
                "Cannot calculate chi squared for fit with "
                f"{len(self.free_parameters)} floated parameters and only "
                f"{len(x)} data points."
            )

        y_fit = self.model.func(x, self.values)
        chi_2 = np.sum(np.power((y - y_fit) / sigma, 2))
        p = stats.chi2.sf(chi_2, n)

        return p

    def _fit_significance(self) -> Optional[float]:
        """Returns an estimate of the goodness of fit as a number between 0 and 1 or
        `None` if `sigma` has not been supplied. See :meth chi_squared: for details.
        """
        if self.sigma is None:
            return None

        return self.chi_squared(self.x, self.y, self.sigma)



from typing import Dict, TYPE_CHECKING
import numpy as np

from .. import Model, ModelParameter
from ..utils import Array

if TYPE_CHECKING:
    num_samples = float


class Rectangle(Model):
    """Rectangle function according to:
    x <= x_l: y = y0
    x >= x_r: y = y0
    x_r > x > x_l: y0 + a

    Fit parameters (all floated by default unless stated otherwise):
      - a: rectangle height above the baseline
      - y0: y-axis offset
      - x_l: left transition point
      - x_r: right transition point

    Derived parameters:
      None

    For `x_l = y0 = 0`, `x_r = inf` this is a Heaviside step function.
    """

    def __init__(self, thresh: float = 0.5):
        """threshold is used to configure the parameter estimator"""
        self.thresh = thresh
        super().__init__()

    def get_num_y_channels(self) -> int:
        """Returns the number of y channels supported by the model"""
        return 1

    # pytype: disable=invalid-annotation
    def _func(
        self,
        x: Array[("num_samples",), np.float64],
        a: ModelParameter(scale_func=lambda x_scale, y_scale, _: y_scale),
        y0: ModelParameter(scale_func=lambda x_scale, y_scale, _: y_scale),
        x_l: ModelParameter(scale_func=lambda x_scale, y_scale, _: x_scale),
        x_r: ModelParameter(scale_func=lambda x_scale, y_scale, _: x_scale),
    ) -> Array[("num_samples",), np.float64]:
        return np.where(np.logical_and(x_r > x, x > x_l), y0 + a, y0)

    # pytype: enable=invalid-annotation

    def estimate_parameters(
        self,
        x: Array[("num_samples",), np.float64],
        y: Array[("num_samples",), np.float64],
        model_parameters: Dict[str, ModelParameter],
    ):
        """Set heuristic values for model parameters.

        Typically called during `Fitter.fit`. This method may make use of information
        supplied by the user for some parameters (via the `fixed_to` or
        `user_estimate` attributes) to find initial guesses for other parameters.

        The datasets must be sorted in order of increasing x-axis values and must not
        contain any infinite or nan values. If all parameters of the model allow
        rescaling, then `x`, `y` and `model_parameters` will contain rescaled values.

        :param x: x-axis data, rescaled if allowed.
        :param y: y-axis data, rescaled if allowed.
        :param model_parameters: dictionary mapping model parameter names to their
            metadata, rescaled if allowed.
        """
        # Ensure that y is a 1D array
        y = np.squeeze(y)

        unknowns = {
            param
            for param, param_data in model_parameters.items()
            if not param_data.has_user_initial_value()
        }

        if {"x_l", "x_r"}.issubset(unknowns):
            model_parameters["y0"].heuristic = 0.5 * (y[0] + y[-1])

        elif "x_l" not in unknowns:
            x_l = model_parameters["x_l"].get_initial_value()

            if min(x) < x_l:
                model_parameters["y0"].heuristic = np.mean(y[x < x_l])
            else:
                y0 = model_parameters["y0"].heuristic = y[-1]
                model_parameters["a"].heuristic = y[0] - y0

        elif "x_r" not in unknowns:
            x_r = model_parameters["x_r"].get_initial_value()
            if max(x) > x_r:
                model_parameters["y0"].heuristic = np.mean(y[x > x_r])
            else:
                y0 = model_parameters["y0"].heuristic = y[0]
                model_parameters["a"].heuristic = y[-1] - y0

        else:
            x_l = model_parameters["x_l"].get_initial_value()
            x_r = model_parameters["x_r"].get_initial_value()

            outside = np.logical_or(x <= x_l, x >= x_r)
            inside = np.logical_and(x > x_l, x < x_r)
            model_parameters["y0"].heuristic = np.mean(y[outside])
            y0 = model_parameters["y0"].get_initial_value()
            model_parameters["a"].heuristic = np.mean(y[outside] - y0)

        y0 = model_parameters["y0"].get_initial_value()
        model_parameters["a"].heuristic = y[np.argmax(np.abs(y - y0))] - y0
        a = model_parameters["a"].get_initial_value()

        thresh = self.thresh * (y0 + (y0 + a))
        inside = (y >= thresh) if a > 0 else (y < thresh)

        if x[inside].size == 0:
            x_l = x[0]
            x_r = x[-1]
        else:
            x_l = min(x[inside])
            x_r = max(x[inside])

        model_parameters["x_l"].heuristic = x_l
        model_parameters["x_r"].heuristic = x_r



from typing import Dict, Tuple, TYPE_CHECKING

import numpy as np

from .. import Model, ModelParameter
from ..utils import Array

if TYPE_CHECKING:
    num_samples = float


def _scale_func(x_scale, y_scale, model):
    # Prevent rescaling
    return None


def _generate_benchmarking_parameters(num_qubits):
    params = {
        "p": ModelParameter(lower_bound=0.0, upper_bound=1.0, scale_func=_scale_func),
        "y0": ModelParameter(
            lower_bound=0.0,
            upper_bound=1.0,
            scale_func=lambda x_scale, y_scale, _: y_scale,
        ),
        "y_inf": ModelParameter(
            fixed_to=1 / 2**num_qubits,
            lower_bound=0,
            upper_bound=1,
            scale_func=lambda x_scale, y_scale, _: y_scale,
        ),
    }
    return params


class Benchmarking(Model):
    """Benchmarking success probability decay model

    y = (y0 - y_inf)*p^x + y_inf
    for sequence length x.

    Fit parameters (all floated by default unless stated otherwise):
      - p: depolarisation parameter
      - y0: SPAM fidelity estimate
      - y_inf: depolarisation offset (y-axis asymptote) (fixed to 1/2^n by default)

    Derived parameters:
      - e: error per Clifford = (1 - p) / alpha_n where alpha_n = 2^n / (2^n - 1)
      - e_spam: estimated SPAM error = 1 - y0
    """

    def __init__(self, num_qubits):
        """Init

        :param num_qubits: The number of qubits involved in the benchmarking sequence.
        """
        self.num_qubits = num_qubits
        self.alpha = 2**num_qubits / (2**num_qubits - 1)
        super().__init__(parameters=_generate_benchmarking_parameters(num_qubits))

    def func(
        self, x: Array[("num_samples",), np.float64], params: Dict[str, float]
    ) -> Array[("num_samples",), np.float64]:
        """Evaluates the model at a given set of x-axis points and with a given
        parameter set and returns the result."""
        p = params["p"]
        y0 = params["y0"]
        y_inf = params["y_inf"]
        y = (y0 - y_inf) * p**x + y_inf
        return y

    def estimate_parameters(
        self,
        x: Array[("num_samples",), np.float64],
        y: Array[("num_samples",), np.float64],
        model_parameters: Dict[str, ModelParameter],
    ):
        """Set heuristic values for model parameters.

        Typically called during `Fitter.fit`. This method may make use of information
        supplied by the user for some parameters (via the `fixed_to` or
        `user_estimate` attributes) to find initial guesses for other parameters.

        The datasets must be sorted in order of increasing x-axis values and must not
        contain any infinite or nan values. If all parameters of the model allow
        rescaling, then `x`, `y` and `model_parameters` will contain rescaled values.

        :param x: x-axis data, rescaled if allowed.
        :param y: y-axis data, rescaled if allowed.
        :param model_parameters: dictionary mapping model parameter names to their
            metadata, rescaled if allowed.
        """
        model_parameters["p"].heuristic = 1.0
        model_parameters["y0"].heuristic = max(y)
        model_parameters["y_inf"].heuristic = 1 / 2**self.num_qubits

    def calculate_derived_params(
        self,
        x: Array[("num_samples",), np.float64],
        y: Array[("num_samples",), np.float64],
        fitted_params: Dict[str, float],
        fit_uncertainties: Dict[str, float],
    ) -> Tuple[Dict[str, float], Dict[str, float]]:
        """
        Returns dictionaries of values and uncertainties for the derived model
        parameters (parameters which are calculated from the fit results rather than
        being directly part of the fit) based on values of the fitted parameters and
        their uncertainties.

        :param x: x-axis data
        :param y: y-axis data
        :param: fitted_params: dictionary mapping model parameter names to their
            fitted values.
        :param fit_uncertainties: dictionary mapping model parameter names to
            their fit uncertainties.
        :returns: tuple of dictionaries mapping derived parameter names to their
            values and uncertainties.
        """
        p = fitted_params["p"]
        y0 = fitted_params["y0"]

        e = (1 - p) / self.alpha
        e_spam = 1 - y0

        p_err = fit_uncertainties["p"]
        y0_err = fit_uncertainties["y0"]

        derived_params = {}
        derived_params["e"] = e
        derived_params["e_spam"] = e_spam

        derived_uncertainties = {}
        derived_uncertainties["e"] = p_err / self.alpha
        derived_uncertainties["e_spam"] = y0_err

        return derived_params, derived_uncertainties



from .containers import AggregateModel, RepeatedModel
from .benchmarking import Benchmarking
from .exponential import Exponential
from .gaussian import Gaussian
from .laser_rabi import (
    LaserFlopFreqCoherent,
    LaserFlopFreqSqueezed,
    LaserFlopFreqThermal,
    LaserFlopTimeCoherent,
    LaserFlopTimeSqueezed,
    LaserFlopTimeThermal,
)
from .lorentzian import Lorentzian
from .polynomial import Power, Polynomial, Line, Parabola
from . import quantum_phys
from .rabi import RabiFlopFreq, RabiFlopTime
from .rectangle import Rectangle
from .sinc import Sinc, Sinc2
from .sinusoid import Sinusoid
from .triangle import Triangle
from . import utils



import copy
import inspect
from typing import Dict, Tuple, TYPE_CHECKING

import numpy as np
from scipy import special

from .quantum_phys import (
    coherent_state_probs,
    squeezed_state_probs,
    thermal_state_probs,
)
from . import rabi
from .. import ModelParameter
from ..utils import Array


if TYPE_CHECKING:
    num_samples = float
    num_fock_states = float


def make_laser_flop(base_class, distribution_fun):
    """
    Generate subclass for laser Rabi flopping models.

    :param base_class: base class to inherit from
    :param distribution_fun: function returning an array of Fock state occupation
        probabilities. The distribution function's first argument should be the maximum
        Fock state to include in the simulation (the returned array has n_max + 1
        elements). Subsequent arguments should be `ModelParameter`s used to parametrise
        the distribution.
    """

    class LaserFlop(base_class):
        """Base class for damped Rabi flopping with finite Lamb-Dicke parameter.

        This model calculates measurement outcomes for systems containing two
        internal states and moving in a 1D harmonic potential that undergo
        damped Rabi oscillations, defined by:
            `P = P_readout_g + (P_readout_e - P_readout_g) * P_e`
        where `P_e` is the (time-dependent) population in the excited state and
        `P_readout_g` and `P_readout_e` are the readout levels (measurement outcomes
        when the qubit is in one state).

        This class does not support fitting directly; use one of the subclasses instead.

        The model requires that the internal part of the system starts out
        entirely in one of the ground or excited states, specified using
        :meth:`__init__`'s :param:`start_excited` parameter. It further assumes
        that the motional part of the system starts out in a distribution
        over different Fock states, described by the parameter `distribution_fun`.

        Independent variables:
            - t_pulse: duration of driving pulse including dead time. The duration of
                the interaction is given by t = max(0, t_pulse - t_dead).
            - w: frequency of driving pulse relative to the reference frequency `w_0`,
                given by `delta = w - w_0`

        Model parameters:
            - P_readout_e: excited state readout level
            - P_readout_g: ground state readout level
            - eta: Lamb-Dicke parameter
            - omega: carrier Rabi frequency
            - tau: decay time constant (fixed to infinity by default)
            - t_dead: dead time (fixed to 0 by default)
            - w_0: resonance frequency offset

        The model additionally gains any parameters associated with the specified Fock
        state distribution function.

        Derived parameters:
            - t_pi: Pi-time, calculated as t_pi = pi / omega
            - t_pi_2: Pi/2-time, calculated as t_pi_2 = t_pi / 2
            - f_0: Offset of resonance from zero of frequency variable in linear units

        All frequencies are in angular units.

        See also models.rabi.RabiFlop.
        """

        def __init__(
            self,
            start_excited: bool,
            sideband_index: int,
            n_max: int = 30,
        ):
            """
            :param start_excited: if True the qubit starts in the excited state
            :param sideband_index: change in motional state due to a pi-pulse starting
              from the spin ground-state.
            :param n_max: maximum Fock state used in the simulation
            """
            super().__init__(start_excited)

            self.distribution_fun = distribution_fun
            self.sideband_index = sideband_index
            self.n_max = n_max

            n = np.arange(self.n_max + 1)
            self._n_min = n_min = np.minimum(n, n + self.sideband_index)
            n_max = np.maximum(n, n + self.sideband_index)

            # sqrt(n_min!/n_max!)
            self.fact = np.exp(
                0.5 * (special.gammaln(n_min + 1) - special.gammaln(n_max + 1))
            )
            self.fact[self._n_min < 0] = 0

            # Add parameters associated with the Fock state distribution function
            spec = inspect.getfullargspec(distribution_fun)
            self.distribution_params = {
                name: copy.deepcopy(spec.annotations[name]) for name in spec.args[1:]
            }

            assert all(
                [
                    isinstance(param, ModelParameter)
                    for param in self.distribution_params.values()
                ]
            ), "Distribution function parameters must be instances of `ModelParameter`"
            assert not any(
                [
                    param_name in self.parameters.keys()
                    for param_name in self.distribution_params.keys()
                ]
            ), "Distribution parameter names must not match model parameter names"
            self.parameters.update(self.distribution_params)

        # pytype: disable=invalid-annotation
        def _func(
            self,
            x: Tuple[
                Array[("num_samples",), np.float64], Array[("num_samples",), np.float64]
            ],
            P_readout_e: ModelParameter(
                lower_bound=0.0,
                upper_bound=1.0,
                scale_func=lambda x_scale, y_scale, _: y_scale,
            ),
            P_readout_g: ModelParameter(
                lower_bound=0.0,
                upper_bound=1.0,
                scale_func=lambda x_scale, y_scale, _: y_scale,
            ),
            eta: ModelParameter(lower_bound=0.0),
            omega: ModelParameter(lower_bound=0.0),
            tau: ModelParameter(lower_bound=0.0, fixed_to=np.inf),
            t_dead: ModelParameter(lower_bound=0.0, fixed_to=0.0),
            w_0: ModelParameter(),
            **kwargs,  # Fock state distribution function parameters
        ) -> Array[("num_samples",), np.float64]:
            """
            Return measurement probability.

            :param x: Tuple (t_pulse, w) of ndarrays containing pulse duration and
                angular frequency of driving field. They must have shapes such
                that they are broadcastable.
            """
            # coupling between |g>|n> and |e>|n+sideband_index>
            omega_vec = (
                omega
                * np.exp(-0.5 * np.power(eta, 2))
                * np.power(eta, np.abs(self.sideband_index))
                * self.fact
                * special.eval_genlaguerre(
                    self._n_min, np.abs(self.sideband_index), np.power(eta, 2)
                )
            )

            t_vec = np.clip(x[0] - t_dead, a_min=0, a_max=None)
            detuning_vec = x[1]
            t, detuning, omega = np.meshgrid(
                t_vec, detuning_vec, omega_vec, indexing="ij"
            )

            # Transition probability for individual Fock state, given by
            #     P_transition = 1/2 * omega^2 / W^2 * [1 - exp(-t / tau) * cos(W * t)]
            W = np.sqrt(np.power(omega, 2) + np.power(detuning, 2))
            P_trans = (
                0.5
                * np.power(
                    np.divide(omega, W, out=np.zeros_like(W), where=(W != 0.0)), 2
                )
                * (1 - np.exp(-t / tau) * np.cos(W * t))
            )

            # Occupation probabilities of Fock states
            P_fock = self.distribution_fun(self.n_max, **kwargs)
            # Transition probability averaged over Fock state distribution
            P_trans_mean = np.sum(P_fock * P_trans, axis=-1).squeeze()

            P_e = 1 - P_trans_mean if self.start_excited else P_trans_mean

            return P_readout_g + (P_readout_e - P_readout_g) * P_e

        # pytype: enable=invalid-annotation

        def estimate_parameters(
            self,
            x: Array[("num_samples",), np.float64],
            y: Array[("num_samples",), np.float64],
            model_parameters: Dict[str, ModelParameter],
        ):
            # Pick sensible starting values which are usually good enough for the fit to
            # converge from.
            model_parameters["eta"].heuristic = 0.1
            if "n_bar" in model_parameters.keys():
                model_parameters["n_bar"].heuristic = 1
            if "alpha" in model_parameters.keys():
                model_parameters["alpha"].heuristic = 1
            if "zeta" in model_parameters.keys():
                model_parameters["zeta"].heuristic = 1

            super().estimate_parameters(x=x, y=y, model_parameters=model_parameters)

    return LaserFlop


def make_laser_flop_freq(distribution_fun):
    """
    Fit model for Rabi flopping pulse detuning scans. See models.rabi.RabiFlopFreq.
    """
    return make_laser_flop(rabi.RabiFlopFreq, distribution_fun)


def make_laser_flop_time(distribution_fun):
    """
    Fit model for laser flopping pulse duration scans. See models.rabi.RabiFlopTime.
    """
    return make_laser_flop(rabi.RabiFlopTime, distribution_fun)


LaserFlopFreqCoherent = make_laser_flop_freq(coherent_state_probs)
LaserFlopFreqThermal = make_laser_flop_freq(thermal_state_probs)
LaserFlopFreqSqueezed = make_laser_flop_freq(squeezed_state_probs)
LaserFlopTimeCoherent = make_laser_flop_time(coherent_state_probs)
LaserFlopTimeThermal = make_laser_flop_time(thermal_state_probs)
LaserFlopTimeSqueezed = make_laser_flop_time(squeezed_state_probs)



import copy
from typing import Dict, List, Optional, Tuple, TYPE_CHECKING

import numpy as np

from .. import Model, ModelParameter
from ..utils import Array


if TYPE_CHECKING:
    num_samples = float
    num_y_channels = float


class AggregateModel(Model):
    """Model formed by aggregating one or more models"""

    def __init__(
        self,
        models: List[Tuple[str, Model]],
        common_params: Optional[List[str]] = None,
    ):
        """
        :param models: The models to be aggregated. This should be a list of tuples,
          with each element containing a model name string and a model instance. The
          model names are used as prefixes for names of model parameters and derived
          results. For example, if one of the aggregated models named `foo` has a
          parameter `bar`, the aggregate model will have a parameter `foo_bar`.

        At present this class only supports models with a single y channel. This is just
        because no one got around to implementing it yet rather than any fundamental
        difficulty.

        A possible future extension of this class would be to introduce a notion of
        "common parameters", whose value is the same for all of the aggregated models. A
        design question here is where the common parameters should get their properties
        (bounds, scale factors, etc) from and how we should deal with their heuristics.
        """
        self.models = copy.deepcopy(models)

        parameters = {}
        for model_name, model in self.models:
            if model.get_num_y_channels() != 1:
                raise ValueError(
                    "AggregateModel only supports models with a single y channel"
                )
            parameters.update(
                {
                    f"{model_name}_{param_name}": param
                    for param_name, param in model.parameters.items()
                }
            )

        super().__init__(parameters=parameters)

    def get_num_y_channels(self) -> int:
        return len(self.models)

    def func(
        self,
        x: Array[("num_samples",), np.float64],
        param_values: Dict[str, float],
    ) -> Array[("num_y_channels", "num_samples"), np.float64]:
        ys = []
        for model_name, model in self.models:
            model_params = {
                param_name: param_values[f"{model_name}_{param_name}"]
                for param_name in model.parameters.keys()
            }
            ys.append(model.func(x, model_params))

        return np.array(ys)

    def estimate_parameters(
        self,
        x: Array[("num_samples",), np.float64],
        y: Array[("num_y_channels", "num_samples"), np.float64],
        model_parameters: Dict[str, ModelParameter],
    ):
        for idx, (model_name, model) in enumerate(self.models):
            params = {
                param_name: model_parameters[f"{model_name}_{param_name}"]
                for param_name in model.parameters.keys()
            }
            model.estimate_parameters(x, y[idx], params)

    def calculate_derived_params(
        self,
        x: Array[("num_samples",), np.float64],
        y: Array[("num_y_channels", "num_samples"), np.float64],
        fitted_params: Dict[str, float],
        fit_uncertainties: Dict[str, float],
    ) -> Tuple[Dict[str, float], Dict[str, float]]:
        derived_params = {}
        derived_uncertainties = {}

        for idx, (model_name, model) in enumerate(self.models):
            model_fitted_params = {
                param_name: fitted_params[f"{model_name}_{param_name}"]
                for param_name in model.parameters.keys()
            }
            model_fitted_uncertainties = {
                param_name: fit_uncertainties[f"{model_name}_{param_name}"]
                for param_name in model.parameters.keys()
            }

            derived = model.calculate_derived_params(
                x=x,
                y=y[idx],
                fitted_params=model_fitted_params,
                fit_uncertainties=model_fitted_uncertainties,
            )
            model_derived_params, model_derived_uncertainties = derived

            derived_params.update(
                {
                    f"{model_name}_{param_name}": value
                    for param_name, value in model_derived_params.items()
                }
            )
            derived_uncertainties.update(
                {
                    f"{model_name}_{param_name}": value
                    for param_name, value in model_derived_uncertainties.items()
                }
            )

        return derived_params, derived_uncertainties


class RepeatedModel(Model):
    """Model formed by repeating a `Model` one more times

    The `RepeatedModel` has multiple y-channels, corresponding to the repetitions of
    the wrapped model.
    """

    def __init__(
        self,
        inner: Model,
        common_params: Optional[List[str]] = None,
        num_repetitions: int = 2,
    ):
        """
        :param inner: The wrapped model, the implementation of `inner` will be used to
          generate data for the y channels
        :param common_params: optional list of names of model arguments, whose value is
          common to all y channels. All other model parameters are independent
        :param num_repetitions: the number of times the inner model is repeated

        Parameters of the new model:
          - all common parameters of the inner model are parameters of the outer model
          - for each independent (not common) or derived parameter of the inner model,
            `foo`, the outer model has parameters `foo_{n}` for n in
            [0, .., num_repitions-1]
          - for each independent or derived parameter, `foo`, the outer model calculates
            statistics for the results from the various models: `foo_mean` and
            `foo_peak_peak`.
        """
        inner = copy.deepcopy(inner)

        inner_params = set(inner.parameters.keys())
        common_params = set(common_params or [])

        if not common_params.issubset(inner_params):
            raise ValueError(
                "Common parameters must be a subset of the inner model's parameters"
            )

        independent_params = set(inner.parameters.keys()) - common_params
        params = {param: inner.parameters[param] for param in common_params}
        for param in independent_params:
            params.update(
                {
                    f"{param}_{idx}": copy.deepcopy(inner.parameters[param])
                    for idx in range(num_repetitions)
                }
            )

        super().__init__(parameters=params)

        self.inner = inner
        self.common_params = common_params
        self.independent_params = independent_params
        self.num_repetitions = num_repetitions

    def get_num_y_channels(self) -> int:
        return self.num_repetitions * self.inner.get_num_y_channels()

    def func(
        self,
        x: Array[("num_samples",), np.float64],
        param_values: Dict[str, float],
    ) -> Array[("num_y_channels", "num_samples"), np.float64]:
        common_values = {param: param_values[param] for param in self.common_params}

        ys = []
        for idx in range(self.num_repetitions):
            values = dict(common_values)
            values.update(
                {
                    param: param_values[f"{param}_{idx}"]
                    for param in self.independent_params
                }
            )
            ys.append(np.atleast_2d(self.inner.func(x, values)))

        return np.vstack(ys)

    def estimate_parameters(
        self,
        x: Array[("num_samples",), np.float64],
        y: Array[("num_y_channels", "num_samples"), np.float64],
        model_parameters: Dict[str, ModelParameter],
    ):
        dim = self.inner.get_num_y_channels()

        common_params = {param: model_parameters[param] for param in self.common_params}
        common_heuristics = {param: [] for param in self.common_params}
        for idx in range(self.num_repetitions):
            params = {
                param: model_parameters[f"{param}_{idx}"]
                for param in self.independent_params
            }
            params.update(copy.deepcopy(common_params))
            self.inner.estimate_parameters(x, y[idx * dim : (idx + 1) * dim], params)

            for param in self.common_params:
                common_heuristics[param].append(params[param].get_initial_value())

        for param in self.common_params:
            model_parameters[param].heuristic = np.mean(common_heuristics[param])

    def calculate_derived_params(
        self,
        x: Array[("num_samples",), np.float64],
        y: Array[("num_samples",), np.float64],
        fitted_params: Dict[str, float],
        fit_uncertainties: Dict[str, float],
    ) -> Tuple[Dict[str, float], Dict[str, float]]:
        derived_params = {}
        derived_uncertainties = {}

        dim = self.inner.get_num_y_channels()

        for idx in range(self.num_repetitions):
            rep_params = {
                param: fitted_params[f"{param}_{idx}"]
                for param in self.independent_params
            }
            rep_params.update(
                {param: fitted_params[param] for param in self.common_params}
            )

            rep_uncertainties = {
                param: fit_uncertainties[f"{param}_{idx}"]
                for param in self.independent_params
            }
            rep_uncertainties.update(
                {param: fit_uncertainties[param] for param in self.common_params}
            )

            derived = self.inner.calculate_derived_params(
                x=x,
                y=y[idx * dim : (idx + 1) * dim],
                fitted_params=rep_params,
                fit_uncertainties=rep_uncertainties,
            )
            rep_derived_params, rep_derived_uncertainties = derived

            derived_params.update(
                {f"{param}_{idx}": value for param, value in rep_derived_params.items()}
            )
            derived_uncertainties.update(
                {
                    f"{param}_{idx}": value
                    for param, value in rep_derived_uncertainties.items()
                }
            )

        # calculate statistical results
        def add_statistics(values, uncertainties, param_names):
            for param_name in param_names:
                param_values = [
                    values[f"{param_name}_{idx}"] for idx in range(self.num_repetitions)
                ]
                param_uncerts = [
                    uncertainties[f"{param_name}_{idx}"]
                    for idx in range(self.num_repetitions)
                ]

                derived_params[f"{param_name}_mean"] = np.mean(param_values)
                derived_uncertainties[f"{param_name}_mean"] = (
                    np.sqrt(np.sum(np.power(param_uncerts, 2))) / self.num_repetitions
                )

                derived_params[f"{param_name}_peak_peak"] = np.ptp(param_values)
                derived_uncertainties[f"{param_name}_peak_peak"] = float("nan")

        add_statistics(
            derived_params, derived_uncertainties, param_names=rep_derived_params.keys()
        )
        add_statistics(
            fitted_params,
            fit_uncertainties,
            param_names=self.independent_params,
        )

        return derived_params, derived_uncertainties


class MappedModel(Model):
    """`Model` wrapping another `Model` with renamed parameters"""

    def __init__(
        self,
        inner: Model,
        mapped_params: Dict[str, str],
        fixed_params: Optional[Dict[str, float]] = None,
    ):
        """Init

        :param inner: The wrapped model, the implementation of `inner` will be used
            after the parameter mapping has been done.
        :param mapped_params: dictionary mapping names of parameters in the new
            model to names of parameters used in the wrapped model.
        :param fixed_params: dictionary mapping names of parameters used in the
            wrapped model to values they are fixed to in the new model. These
            will not be parameters of the new model.
        """
        inner_params = inner.parameters

        if fixed_params is None:
            fixed_params = {}

        if unknown_mapped_params := set(mapped_params.values()) - inner_params.keys():
            raise ValueError(
                "The target of parameter mappings must be parameters of the inner "
                f"model. The mapping targets are not: {unknown_mapped_params}"
            )

        if unknown_fixed_params := fixed_params.keys() - inner_params.keys():
            raise ValueError(
                "Fixed parameters must be parameters of the inner model. The "
                f"follow fixed parameters are not: {unknown_fixed_params}"
            )

        if missing_params := inner_params.keys() - (
            fixed_params.keys() | mapped_params.values()
        ):
            raise ValueError(
                "All parameters of the inner model must be either mapped of "
                "fixed. The following inner model parameters are neither: "
                f"{missing_params}"
            )

        if duplicated_params := fixed_params.keys() & mapped_params.values():
            raise ValueError(
                "Parameters cannot be both mapped and fixed. The following "
                f"parameters are both: {duplicated_params}"
            )

        params = {
            new_name: inner_params[old_name]
            for new_name, old_name in mapped_params.items()
        }
        super().__init__(parameters=params)
        self.inner = inner
        self.mapped_args = mapped_params
        self.fixed_params = fixed_params or {}

    def can_rescale(self, x_scale: float, y_scale: float) -> bool:
        """Returns True if the model can be rescaled"""
        return self.inner.can_rescale(x_scale, y_scale)

    @staticmethod
    def get_scaled_model(model, x_scale: float, y_scale: float):
        """Returns a scaled copy of a given model object

        :param model: model to be copied and rescaled
        :param x_scale: x-axis scale factor
        :param y_scale: y-axis scale factor
        :returns: a scaled copy of model
        """
        scaled_model = copy.deepcopy(model)
        for param_name, param in scaled_model.inner.parameters.items():
            param.rescale(x_scale, y_scale, scaled_model.inner)

        for fixed_param in scaled_model.fixed_params.keys():
            scale_factor = scaled_model.inner.parameters[fixed_param].scale_factor
            scaled_model.fixed_params[fixed_param] /= scale_factor

        # Expose the scale factors to the fitter so it knows how to rescale the results
        for new_name, old_name in scaled_model.mapped_args.items():
            scale_factor = scaled_model.inner.parameters[old_name].scale_factor
            scaled_model.parameters[new_name].scale_factor = scale_factor

        return scaled_model

    def get_num_y_channels(self) -> int:
        """Returns the number of y channels supported by the model"""
        return self.inner.get_num_y_channels()

    def func(
        self, x: Array[("num_samples",), np.float64], param_values: Dict[str, float]
    ) -> Array[("num_samples", "num_y_channels"), np.float64]:
        """Evaluates the model at a given set of x-axis points and with a given set of
        parameter values and returns the result.

        Overload this to provide a model function with a dynamic set of parameters,
        otherwise prefer to override `_func`.

        :param x: x-axis data
        :param param_values: dictionary of parameter values
        :returns: array of model values
        """
        new_params = {
            old_name: param_values[new_name]
            for new_name, old_name in self.mapped_args.items()
        }
        new_params.update(self.fixed_params)
        return self.inner.func(x, new_params)

    def _inner_estimate_parameters(
        self,
        x: Array[("num_samples",), np.float64],
        y: Array[("num_samples", "num_y_channels"), np.float64],
        inner_parameters: Dict[str, ModelParameter],
    ) -> Dict[str, float]:
        return self.inner.estimate_parameters(x, y, inner_parameters)

    def estimate_parameters(
        self,
        x: Array[("num_samples",), np.float64],
        y: Array[("num_samples", "num_y_channels"), np.float64],
        model_parameters: Dict[str, ModelParameter],
    ):
        """Set heuristic values for model parameters.

        Typically called during `Fitter.fit`. This method may make use of information
        supplied by the user for some parameters (via the `fixed_to` or
        `user_estimate` attributes) to find initial guesses for other parameters.

        The datasets must be sorted in order of increasing x-axis values and must not
        contain any infinite or nan values. If all parameters of the model allow
        rescaling, then `x`, `y` and `model_parameters` will contain rescaled values.

        :param x: x-axis data, rescaled if allowed.
        :param y: y-axis data, rescaled if allowed.
        :param model_parameters: dictionary mapping model parameter names to their
            metadata, rescaled if allowed.
        """
        inner_parameters = {
            original_param: copy.deepcopy(model_parameters[new_param])
            for new_param, original_param in self.mapped_args.items()
        }

        inner_parameters.update(
            {
                param: ModelParameter(
                    lower_bound=value, upper_bound=value, fixed_to=value
                )
                for param, value in self.fixed_params.items()
            }
        )

        self._inner_estimate_parameters(x, y, inner_parameters)

        for new_param, original_param in self.mapped_args.items():
            initial_value = inner_parameters[original_param].get_initial_value()
            model_parameters[new_param].heuristic = initial_value



from typing import Dict, Tuple, TYPE_CHECKING
import numpy as np

from .utils import get_spectrum
from .. import Model, ModelParameter
from ..utils import Array


if TYPE_CHECKING:
    num_samples = float


class Gaussian(Model):
    """Gaussian model according to:
    y = a / (sigma * sqrt(2*pi)) * exp(-(x-x0)^2/(2*sigma)^2) + y0

    Fit parameters (all floated by default unless stated otherwise):
      - x0: x-axis offset
      - y0: y-axis offset
      - a: y-axis scale factor. The Gaussian is normalized such that its integral is
        equal to a.
      - sigma: distribution half-width at 1/e of maximum height is 2*sigma (sigma is the
        1/sqrt(e) radius).

    Derived parameters:
      - FWHMH: full width at half-maximum height
      - peak: peak height above y0
      - w0: full width at 1/e max height. For Gaussian beams this is the beam waist
    """

    def get_num_y_channels(self) -> int:
        """Returns the number of y channels supported by the model"""
        return 1

    # pytype: disable=invalid-annotation
    def _func(
        self,
        x: Array[("num_samples",), np.float64],
        x0: ModelParameter(scale_func=lambda x_scale, y_scale, _: x_scale),
        y0: ModelParameter(scale_func=lambda x_scale, y_scale, _: y_scale),
        a: ModelParameter(scale_func=lambda x_scale, y_scale, _: y_scale * x_scale),
        sigma: ModelParameter(
            lower_bound=0, scale_func=lambda x_scale, y_scale, _: x_scale
        ),
    ) -> Array[("num_samples",), np.float64]:
        y = (
            a
            / (sigma * np.sqrt(2 * np.pi))
            * np.exp(-np.power((x - x0) / (2 * sigma), 2))
            + y0
        )
        return y

    # pytype: enable=invalid-annotation

    def estimate_parameters(
        self,
        x: Array[("num_samples",), np.float64],
        y: Array[("num_samples",), np.float64],
        model_parameters: Dict[str, ModelParameter],
    ):
        """Set heuristic values for model parameters.

        Typically called during `Fitter.fit`. This method may make use of information
        supplied by the user for some parameters (via the `fixed_to` or
        `user_estimate` attributes) to find initial guesses for other parameters.

        The datasets must be sorted in order of increasing x-axis values and must not
        contain any infinite or nan values. If all parameters of the model allow
        rescaling, then `x`, `y` and `model_parameters` will contain rescaled values.

        :param x: x-axis data, rescaled if allowed.
        :param y: y-axis data, rescaled if allowed.
        :param model_parameters: dictionary mapping model parameter names to their
            metadata, rescaled if allowed.
        """
        # Ensure that y is a 1D array
        y = np.squeeze(y)

        # Gaussian Fourier Transform:
        #   F[A * exp(-(x/w)^2)](k) = A * sqrt(pi) * w * exp(-(pi*k*w)^2)
        #
        # Half-width at 1/e when k = 1/(pi*w)
        omega, spectrum = get_spectrum(x, y, trim_dc=True)
        abs_spectrum = np.abs(spectrum)

        k = omega / (2 * np.pi)

        peak = np.max(abs_spectrum)
        W = peak / np.exp(1)
        width = 1 / (np.pi * k[np.argmin(np.abs(abs_spectrum - W))])

        sigma = width / 2
        a = peak * np.pi * np.sqrt(2)

        model_parameters["y0"].heuristic = np.mean([y[0], y[-1]])
        y0 = model_parameters["y0"].get_initial_value()
        peak_idx = np.argmax(np.abs(y - y0))
        y_peak = y[peak_idx]
        sgn = 1 if y_peak > y0 else -1

        model_parameters["a"].heuristic = a * sgn
        model_parameters["sigma"].heuristic = sigma

        cut_off = 2 * omega[np.argmin(np.abs(abs_spectrum - W))]

        x0 = self.find_x_offset_sym_peak(
            x=x,
            y=y,
            parameters=model_parameters,
            omega=omega,
            spectrum=spectrum,
            omega_cut_off=cut_off,
            test_pts=x[peak_idx],
        )

        model_parameters["x0"].heuristic = x0

    def calculate_derived_params(
        self,
        x: Array[("num_samples",), np.float64],
        y: Array[("num_samples",), np.float64],
        fitted_params: Dict[str, float],
        fit_uncertainties: Dict[str, float],
    ) -> Tuple[Dict[str, float], Dict[str, float]]:
        """
        Returns dictionaries of values and uncertainties for the derived model
        parameters (parameters which are calculated from the fit results rather than
        being directly part of the fit) based on values of the fitted parameters and
        their uncertainties.

        :param x: x-axis data
        :param y: y-axis data
        :param: fitted_params: dictionary mapping model parameter names to their
            fitted values.
        :param fit_uncertainties: dictionary mapping model parameter names to
            their fit uncertainties.
        :returns: tuple of dictionaries mapping derived parameter names to their
            values and uncertainties.
        """
        sigma = fitted_params["sigma"]
        a = fitted_params["a"]

        derived_params = {}
        derived_params["FWHMH"] = 2.35482 * sigma
        derived_params["peak"] = a / (sigma * np.sqrt(2 * np.pi))
        derived_params["w0"] = 4 * sigma

        derived_uncertainties = {}
        derived_uncertainties["FWHMH"] = 2.35482 * fit_uncertainties["sigma"]
        derived_uncertainties["peak"] = (
            fit_uncertainties["a"] / (sigma * np.sqrt(2 * np.pi))
        ) ** 2 + (
            fit_uncertainties["sigma"] * a / (sigma**2 * np.sqrt(2 * np.pi))
        ) ** 2
        derived_uncertainties["w0"] = 4 * fit_uncertainties["sigma"]

        return derived_params, derived_uncertainties



from typing import Dict, Tuple, TYPE_CHECKING

import numpy as np

from .sinc import Sinc2
from .sinusoid import Sinusoid
from .. import Model, ModelParameter, NormalFitter
from .utils import get_spectrum
from ..utils import Array


if TYPE_CHECKING:
    num_samples = float


class RabiFlop(Model):
    """Base class for damped Rabi flopping.

    This model calculates measurement outcomes for two-state systems undergoing damped
    Rabi oscillations, defined by:
        `P = P_readout_g + (P_readout_e - P_readout_g) * P_e`
    where `P_e` is the (time-dependent) population in the excited state and
    `P_readout_g` and `P_readout_e` are the readout levels (measurement outcomes
    when the qubit is in one state).

    This class does not support fitting directly; use one of the subclasses instead.

    The model requires that the system starts out entirely in one of the ground or
    excited states, specified using :meth:`__init__`'s :param:`start_excited` parameter.

    The probability of transition from one state to the other is calculated as
        `P_trans = 1 / 2 * omega^2 / W^2 * [1 - exp(-t / tau) * cos(W * t)]`
    where
        - t is the duration of interaction between qubit and driving field
        - W = sqrt(omega^2 + delta^2)
        - delta is the detuning of the driving field from resonance
        - omega is the Rabi frequency
        - tau is the decay time constant.

    Independent variables:
        - t_pulse: duration of driving pulse including dead time. The duration of the
            interaction is given by t = max(0, t_pulse - t_dead).
        - w: frequency of driving pulse relative to the reference frequency `w_0`, given
            by `delta = w - w_0`

    Model parameters:
        - P_readout_e: excited state readout level
        - P_readout_g: ground state readout level
        - omega: Rabi frequency
        - tau: decay time constant (fixed to infinity by default)
        - t_dead: dead time (fixed to 0 by default)
        - w_0: resonance frequency offset

    Derived parameters:
        - t_pi: Pi-time, calculated as t_pi = pi / omega
        - t_pi_2: Pi/2-time, calculated as t_pi_2 = t_pi / 2
        - f_0: Offset of resonance from zero of frequency variable in linear units

    All frequencies are in angular units.
    """

    def __init__(self, start_excited: bool):
        super().__init__()
        self.start_excited = start_excited

    def get_num_y_channels(self) -> int:
        """Returns the number of y channels supported by the model"""
        return 1

    # pytype: disable=invalid-annotation
    def _func(
        self,
        # Beware if you're sub-typing this!
        # This is not the standard type for `x`; we rely on the implementation of `func`
        # to change the type of `x` for us (see the RabiFlopFreq / RabiFlopTime
        # implementations)
        x: Tuple[
            Array[("num_samples",), np.float64], Array[("num_samples",), np.float64]
        ],
        P_readout_e: ModelParameter(
            lower_bound=0.0,
            upper_bound=1.0,
            scale_func=lambda x_scale, y_scale, _: y_scale,
        ),
        P_readout_g: ModelParameter(
            lower_bound=0.0,
            upper_bound=1.0,
            scale_func=lambda x_scale, y_scale, _: y_scale,
        ),
        omega: ModelParameter(lower_bound=0.0),
        tau: ModelParameter(lower_bound=0.0, fixed_to=np.inf),
        t_dead: ModelParameter(lower_bound=0.0, fixed_to=0.0),
        w_0: ModelParameter(),
    ) -> Array[("num_samples",), np.float64]:
        """:param x: tuple of (t_pulse, w)"""
        t = np.clip(x[0] - t_dead, a_min=0.0, a_max=None)
        delta = x[1] - w_0
        W = np.sqrt(omega**2 + delta**2)

        P_trans = (
            0.5
            * np.divide(omega**2, W**2, out=np.zeros_like(W), where=(W != 0.0))
            * (1 - np.exp(-t / tau) * np.cos(W * t))
        )
        P_e = 1 - P_trans if self.start_excited else P_trans

        return P_readout_g + (P_readout_e - P_readout_g) * P_e

    # pytype: enable=invalid-annotation

    def calculate_derived_params(
        self,
        x: Array[("num_samples",), np.float64],
        y: Array[("num_samples",), np.float64],
        fitted_params: Dict[str, float],
        fit_uncertainties: Dict[str, float],
    ) -> Tuple[Dict[str, float], Dict[str, float]]:
        omega = fitted_params["omega"]
        t_pi = np.pi / omega
        t_pi_2 = t_pi / 2

        omega_err = fit_uncertainties["omega"]
        t_dead_err = fit_uncertainties["t_dead"]

        derived_params = {}
        derived_params["t_pi"] = t_pi
        derived_params["t_pi_2"] = t_pi_2

        derived_uncertainties = {}
        derived_uncertainties["t_pi"] = np.sqrt(
            t_dead_err**2 + (omega_err * np.pi / (omega**2)) ** 2
        )
        derived_uncertainties["t_pi_2"] = np.sqrt(
            t_dead_err**2 + (omega_err * np.pi / 2 * (omega**2)) ** 2
        )

        if "w_0" in fitted_params:
            derived_params["f_0"] = fitted_params["w_0"] / (2 * np.pi)
            derived_uncertainties["f_0"] = fit_uncertainties["w_0"] / (2 * np.pi)

        return derived_params, derived_uncertainties


class RabiFlopFreq(RabiFlop):
    """Fit model for Rabi flopping pulse detuning scans.

    This model calculates the measurement outcomes for damped Rabi flops when the
    pulse duration is kept fixed and only its frequency is varied. The pulse duration is
    specified using a new `t_pulse` model parameter.
    """

    def __init__(self, start_excited: bool):
        super().__init__(start_excited)

        self.parameters["t_pulse"] = ModelParameter(lower_bound=0.0)

        self.parameters["t_pulse"].scale_func = lambda x_scale, y_scale, _: 1 / x_scale
        self.parameters["omega"].scale_func = lambda x_scale, y_scale, _: x_scale
        self.parameters["tau"].scale_func = lambda x_scale, y_scale, _: x_scale
        self.parameters["t_dead"].scale_func = lambda x_scale, y_scale, _: 1 / x_scale
        self.parameters["w_0"].scale_func = lambda x_scale, y_scale, _: x_scale

    def func(
        self, x: Array[("num_samples",), np.float64], param_values: Dict[str, float]
    ) -> Array[("num_samples",), np.float64]:
        """Returns measurement outcome as function of pulse frequency.

        :param x: Angular frequency
        """
        param_values = param_values.copy()
        t_pulse = param_values.pop("t_pulse")
        return self._func(
            (t_pulse, x), **param_values
        )  # pytype: disable=wrong-arg-types

    def estimate_parameters(
        self,
        x: Array[("num_samples",), np.float64],
        y: Array[("num_samples",), np.float64],
        model_parameters: Dict[str, ModelParameter],
    ):
        # Ensure that y is a 1D array
        y = np.squeeze(y)

        model_parameters["t_dead"].heuristic = 0.0
        model_parameters["tau"].heuristic = np.inf

        if self.start_excited:
            model_parameters["P_readout_e"].heuristic = y[0]
            model_parameters["P_readout_g"].heuristic = abs(1 - y[0])
        else:
            model_parameters["P_readout_g"].heuristic = y[0]
            model_parameters["P_readout_e"].heuristic = abs(1 - y[0])

        y0_param = "P_readout_e" if self.start_excited else "P_readout_g"

        # A common use of this model is finding `w_0` when the Rabi frequency and pulse
        # duration are known. In this case we don't need to rely on the Sinc2
        # approximation
        unknowns = set()
        for param, param_data in model_parameters.items():
            try:
                param_data.get_initial_value()
            except ValueError:
                unknowns.add(param)

        if unknowns == {"w_0"}:
            omega, spectrum = get_spectrum(x, y, trim_dc=True)
            w_0 = self.find_x_offset_sym_peak(
                x=x,
                y=y,
                parameters=model_parameters,
                omega=omega,
                spectrum=spectrum,
                omega_cut_off=model_parameters["t_pulse"].get_initial_value(),
                x_offset_param_name="w_0",
                y_offset_param_name=y0_param,
            )
            model_parameters["w_0"].heuristic = w_0
            return

        # There isn't a simple analytic form for the Fourier transform of a Rabi
        # flop in the general case. However in the low pulse area limit (and
        # ignoring decay etc) the Rabi flop function tends to the sinc^2 function:
        #   (omega * t / 2)^2 * sinc^2(delta * t / 2)
        # NB np.sinc(x) = np.sin(pi * x) / (pi * x)
        # This heuristic breaks down when: omega * t_pulse ~ pi
        model = Sinc2()
        y0 = model_parameters[y0_param].get_initial_value()
        model.parameters["y0"].fixed_to = y0
        fit = NormalFitter(x, y, model)

        model_parameters["t_pulse"].heuristic = 2 * fit.values["w"]
        t_pulse = model_parameters["t_pulse"].get_initial_value()
        model_parameters["omega"].heuristic = (
            2 * np.sqrt(np.abs(fit.values["a"])) / t_pulse
        )

        if model_parameters["w_0"].has_user_initial_value():
            return

        # The user hasn't told us what w_0 is so we need to find a heuristic value
        # In addition to going off the Sinc^2, we use a simple sampling-based heuristic
        x_sinc = fit.values["x0"]

        # Test out all points with a contrast of 30% of more. NB the fitter
        # automatically rescales our y-data so this assumes we have one point at
        # sufficiently high contrast for the y-axis rescaling to not do much!
        x_sample = x[np.argwhere(np.abs(y - y0) > 0.3)]
        x_trial = np.append(x_sample, [x_sinc])
        w_0, _ = self.param_min_sqrs(
            x, y, model_parameters, scanned_param="w_0", scanned_param_values=x_trial
        )
        model_parameters["w_0"].heuristic = w_0


class RabiFlopTime(RabiFlop):
    """Fit model for Rabi flopping pulse duration scans.

    This model calculates the measurement outcome for damped Rabi flops when the
    frequency of the pulse is kept fixed and only its duration is varied.

    Since the detuning is not scanned as an independent variable, we replace `w_0` with
    a new model parameter `delta`, defined by: delta = |w - w_0|.
    """

    def __init__(self, start_excited: bool):
        super().__init__(start_excited)

        self.parameters["delta"] = ModelParameter()
        del self.parameters["w_0"]

        self.parameters["delta"].scale_func = lambda x_scale, y_scale, _: 1 / x_scale
        self.parameters["omega"].scale_func = lambda x_scale, y_scale, _: 1 / x_scale
        self.parameters["tau"].scale_func = lambda x_scale, y_scale, _: 1 / x_scale
        self.parameters["t_dead"].scale_func = lambda x_scale, y_scale, _: x_scale

    def func(
        self, x: Array[("num_samples",), np.float64], param_values: Dict[str, float]
    ) -> Array[("num_samples",), np.float64]:
        """Returns measurement outcome as function of pulse duration.

        :param x: Pulse duration
        """
        param_values = param_values.copy()
        delta = param_values.pop("delta")
        param_values["w_0"] = 0.0
        return self._func((x, delta), **param_values)  # pytype: disable=wrong-arg-types

    def estimate_parameters(
        self,
        x: Array[("num_samples",), np.float64],
        y: Array[("num_samples",), np.float64],
        model_parameters: Dict[str, ModelParameter],
    ):
        # Ensure that y is a 1D array
        y = np.squeeze(y)

        model_parameters["t_dead"].heuristic = 0.0
        model_parameters["tau"].heuristic = np.inf

        if self.start_excited:
            model_parameters["P_readout_e"].heuristic = y[0]
            model_parameters["P_readout_g"].heuristic = abs(1 - y[0])
        else:
            model_parameters["P_readout_g"].heuristic = y[0]
            model_parameters["P_readout_e"].heuristic = abs(1 - y[0])

        P_readout_e = model_parameters["P_readout_e"].get_initial_value()
        P_readout_g = model_parameters["P_readout_g"].get_initial_value()

        model = Sinusoid()
        if P_readout_e >= P_readout_g:
            model.parameters["phi"].fixed_to = (
                np.pi / 2 if self.start_excited else 3 * np.pi / 2
            )
        else:
            model.parameters["phi"].fixed_to = (
                3 * np.pi / 2 if self.start_excited else np.pi / 2
            )

        fit = NormalFitter(x, y, model)
        W = fit.values["omega"]
        model_parameters["omega"].heuristic = np.sqrt(2 * fit.values["a"]) * W
        omega = model_parameters["omega"].get_initial_value()

        if W >= omega:
            model_parameters["delta"].heuristic = np.sqrt(W**2 - omega**2)
        else:
            # can't use param_min_sqrs because omega and delta are coupled
            deltas = np.linspace(0, W / 2, 10)
            omegas = np.sqrt(W**2 - np.power(deltas, 2))
            costs = np.zeros_like(deltas)

            initial_values = {
                param: param_data.get_initial_value()
                for param, param_data in model_parameters.items()
                if param != "delta"
            }

            for idx in range(len(deltas)):
                initial_values["delta"] = deltas[idx]
                initial_values["omega"] = omegas[idx]
                y_idx = self.func(x, initial_values)
                costs[idx] = np.sqrt(np.sum(np.power(y - y_idx, 2)))
            opt_idx = np.argmin(costs)

            model_parameters["delta"].heuristic = deltas[opt_idx]
            model_parameters["omega"].heuristic = omegas[opt_idx]



import copy
import numpy as np
from typing import Dict, Tuple, TYPE_CHECKING

from .. import Model, ModelParameter
from ..utils import Array
from . import utils

if TYPE_CHECKING:
    num_samples = float


class Sinusoid(Model):
    """Generalised sinusoid fit according to:
    y = Gamma * a * sin[omega * (x - x0) + phi] + y0
    where Gamma = exp(-x / tau).

    Fit parameters (all floated by default unless stated otherwise):
      - a: initial (x = 0) amplitude of the sinusoid
      - omega: angular frequency
      - phi: phase offset (radians)
      - y0: y-axis offset
      - x0: x-axis offset (fixed to 0 by default)
      - tau: decay/growth constant (fixed to np.inf by default)

    Derived parameters:
      - f: frequency
      - phi_cosine: cosine phase (phi + pi/2)
      - contrast: peak-to-peak amplitude of the pure sinusoid
      - min/max: min / max values of the pure sinusoid
      - period: period of oscillation
      - TODO: peak values of the damped sinusoid as well as `x` value that the peak
          occurs at.

    All phases are in radians, frequencies are in angular units.

    x0 and phi0 are equivalent parametrisations for the phase offset, but in some cases
    it works out convenient to have access to both (e.g. one as a fixed offset, the
    other floated). At most one of them should be floated at once. By default, x0 is
    fixed at 0 and phi0 is floated.
    """

    def get_num_y_channels(self) -> int:
        """Returns the number of y channels supported by the model"""
        return 1

    # pytype: disable=invalid-annotation
    def _func(
        self,
        x: Array[("num_samples",), np.float64],
        a: ModelParameter(
            lower_bound=0, scale_func=lambda x_scale, y_scale, _: y_scale
        ),
        omega: ModelParameter(
            lower_bound=0, scale_func=lambda x_scale, y_scale, _: 1 / x_scale
        ),
        phi: utils.PeriodicModelParameter(
            period=2 * np.pi,
            offset=-np.pi,
        ),
        y0: ModelParameter(scale_func=lambda x_scale, y_scale, _: y_scale),
        x0: ModelParameter(fixed_to=0, scale_func=lambda x_scale, y_scale, _: x_scale),
        tau: ModelParameter(
            lower_bound=0,
            fixed_to=np.inf,
            scale_func=lambda x_scale, y_scale, _: x_scale,
        ),
    ) -> Array[("num_samples",), np.float64]:
        Gamma = np.exp(-x / tau)
        y = Gamma * a * np.sin(omega * (x - x0) + phi) + y0
        return y

    # pytype: enable=invalid-annotation

    def estimate_parameters(
        self,
        x: Array[("num_samples",), np.float64],
        y: Array[("num_samples",), np.float64],
        model_parameters: Dict[str, ModelParameter],
    ):
        """Set heuristic values for model parameters.

        Typically called during `Fitter.fit`. This method may make use of information
        supplied by the user for some parameters (via the `fixed_to` or
        `user_estimate` attributes) to find initial guesses for other parameters.

        The datasets must be sorted in order of increasing x-axis values and must not
        contain any infinite or nan values. If all parameters of the model allow
        rescaling, then `x`, `y` and `model_parameters` will contain rescaled values.

        :param x: x-axis data, rescaled if allowed.
        :param y: y-axis data, rescaled if allowed.
        :param model_parameters: dictionary mapping model parameter names to their
            metadata, rescaled if allowed.
        """
        # Ensure that y is a 1D array
        y = np.squeeze(y)

        # We don't have good heuristics for these parameters
        model_parameters["y0"].heuristic = np.mean(y)
        model_parameters["tau"].heuristic = np.max(x)

        omega, spectrum = utils.get_spectrum(x, y, density_units=False, trim_dc=True)
        spectrum = np.abs(spectrum)
        peak = np.argmax(spectrum)

        model_parameters["a"].heuristic = spectrum[peak] * 2
        model_parameters["omega"].heuristic = omega[peak]

        phi_params = copy.deepcopy(model_parameters)
        phi_params["x0"].heuristic = 0.0
        phi, _ = self.param_min_sqrs(
            x=x,
            y=y,
            parameters=phi_params,
            scanned_param="phi",
            scanned_param_values=np.linspace(-np.pi, np.pi, num=20),
        )
        phi = model_parameters["phi"].clip(phi)

        if model_parameters["x0"].fixed_to is None:
            if model_parameters["phi"].fixed_to is None:
                raise ValueError("Only one of 'x0' and 'phi' may be floated at once")

            model_parameters["phi"].heuristic = 0
            model_parameters["x0"].heuristic = (
                -phi / model_parameters["omega"].get_initial_value()
            )
        else:
            model_parameters["phi"].heuristic = phi
            model_parameters["x0"].heuristic = 0.0

    def calculate_derived_params(
        self,
        x: Array[("num_samples",), np.float64],
        y: Array[("num_samples",), np.float64],
        fitted_params: Dict[str, float],
        fit_uncertainties: Dict[str, float],
    ) -> Tuple[Dict[str, float], Dict[str, float]]:
        """
        Returns dictionaries of values and uncertainties for the derived model
        parameters (parameters which are calculated from the fit results rather than
        being directly part of the fit) based on values of the fitted parameters and
        their uncertainties.

        :param x: x-axis data
        :param y: y-axis data
        :param: fitted_params: dictionary mapping model parameter names to their
            fitted values.
        :param fit_uncertainties: dictionary mapping model parameter names to
            their fit uncertainties.
        :returns: tuple of dictionaries mapping derived parameter names to their
            values and uncertainties.
        """
        derived_params = {}
        derived_params["f"] = fitted_params["omega"] / (2 * np.pi)
        derived_params["phi_cosine"] = fitted_params["phi"] + np.pi / 2
        derived_params["contrast"] = 2 * np.abs(fitted_params["a"])
        derived_params["min"] = fitted_params["y0"] - np.abs(fitted_params["a"])
        derived_params["max"] = fitted_params["y0"] + np.abs(fitted_params["a"])
        derived_params["period"] = 2 * np.pi / fitted_params["omega"]

        derived_uncertainties = {}
        derived_uncertainties["f"] = fit_uncertainties["omega"] / (2 * np.pi)
        derived_uncertainties["phi_cosine"] = fit_uncertainties["phi"]
        derived_uncertainties["contrast"] = 2 * fit_uncertainties["a"]
        derived_uncertainties["min"] = np.sqrt(
            fit_uncertainties["y0"] ** 2 + fit_uncertainties["a"] ** 2
        )
        derived_uncertainties["max"] = np.sqrt(
            fit_uncertainties["y0"] ** 2 + fit_uncertainties["a"] ** 2
        )
        derived_uncertainties["period"] = (
            2 * np.pi * fit_uncertainties["omega"] / (fitted_params["omega"] ** 2)
        )

        return derived_params, derived_uncertainties



from typing import Dict, TYPE_CHECKING
import numpy as np

from .exponential import Exponential
from .utils import get_spectrum
from .. import NormalFitter, Model, ModelParameter
from ..utils import Array


if TYPE_CHECKING:
    num_samples = float


class Lorentzian(Model):
    """Lorentzian model according to:
    y = a * fwhmh^2 / ((x - x0)^2 + fwhmh^2) + y0

    Fit parameters (all floated by default unless stated otherwise):
      - x0: x-axis offset
      - y0: y-axis offset
      - a: peak value of the function above y0
      - fwhmh: full width at half maximum height of the function

    Derived parameters:
        None
    """

    def get_num_y_channels(self) -> int:
        """Returns the number of y channels supported by the model"""
        return 1

    # pytype: disable=invalid-annotation
    def _func(
        self,
        x: Array[("num_samples",), np.float64],
        x0: ModelParameter(scale_func=lambda x_scale, y_scale, _: x_scale),
        y0: ModelParameter(scale_func=lambda x_scale, y_scale, _: y_scale),
        a: ModelParameter(scale_func=lambda x_scale, y_scale, _: y_scale),
        fwhmh: ModelParameter(
            lower_bound=0, scale_func=lambda x_scale, y_scale, _: x_scale
        ),
    ) -> Array[("num_samples",), np.float64]:
        y = a * fwhmh**2 / ((x - x0) ** 2 + fwhmh**2) + y0
        return y

    # pytype: enable=invalid-annotation

    def estimate_parameters(
        self,
        x: Array[("num_samples",), np.float64],
        y: Array[("num_samples",), np.float64],
        model_parameters: Dict[str, ModelParameter],
    ):
        """Sets initial values for model parameters based on heuristics. Typically
        called during `Fitter.fit`.

        Heuristic results should be stored in :param model_parameters: using the
        `ModelParameter`'s `heuristic` attribute. This ensures that all information
        passed in by the user (fixed values, initial values, bounds) is used correctly.

        The dataset must be sorted in order of increasing x-axis values and must not
        contain any infinite or nan values.

        :param x: x-axis data
        :param y: y-axis data
        :param model_parameters: dictionary mapping model parameter names to their
            metadata.
        """
        # Ensure that y is a 1D array
        y = np.squeeze(y)

        omega, spectrum = get_spectrum(x, y, trim_dc=True)
        abs_spectrum = np.abs(spectrum)

        model = Exponential()
        model.parameters["y0"].heuristic = np.max(abs_spectrum)
        model.parameters["y_inf"].heuristic = 0.0
        fit = NormalFitter(omega, abs_spectrum, model)

        model_parameters["y0"].heuristic = np.mean([y[0], y[-1]])
        y0 = model_parameters["y0"].get_initial_value()

        peak_idx = np.argmax(np.abs(y - y0))
        y_peak = y[peak_idx]
        sgn = 1 if y_peak > y0 else -1

        model_parameters["fwhmh"].heuristic = 1 / fit.values["tau"]
        fwhmh = model_parameters["fwhmh"].get_initial_value()
        model_parameters["a"].heuristic = fit.values["y0"] * sgn * 2 / fwhmh

        cut_off = 2 * fit.values["tau"]

        x0 = self.find_x_offset_sym_peak(
            x=x,
            y=y,
            parameters=model_parameters,
            omega=omega,
            spectrum=spectrum,
            omega_cut_off=cut_off,
            test_pts=x[peak_idx],
        )

        model_parameters["x0"].heuristic = x0



from typing import Dict, Optional, TYPE_CHECKING

import numpy as np

from .containers import MappedModel
from .. import Model, ModelParameter
from ..utils import Array

if TYPE_CHECKING:
    num_samples = float


def _power_x_scale(x_scale: float, y_scale: float, model: Model) -> Optional[float]:
    if model.parameters["n"].fixed_to is None:
        return None
    return y_scale / np.float_power(x_scale, model.parameters["n"].fixed_to)


class Power(Model):
    """Single-power fit according to:
    y = a*(x-x0)^n + y0

    `x - x0` must always be strictly greater than 0. This is because `n` can take
    non-integral values (for integer coefficients use :class Polynomial: instead)
    and this function's return is real-valued.

    The fit will often struggle when both y0 and n are floated if the dataset doesn't
    contain some asymptotic values where `y ~ y0`. The more you can help it out by
    bounding parameters and providing initial guesses the better.

    The fit will generally struggle to converge if both `a` and `y0` are floated unless
    it is given some guidance (e.g. initial values).

    Fit parameters (all floated by default unless stated otherwise):
      - a: y-axis scale factor (fixed to 1 by default)
      - x0: x-axis offset (fixed to 0 by default). This parameter is not expected to be
        used often.
      - y0: y-axis offset
      - n: power

    Derived parameters:
        None
    """

    def get_num_y_channels(self) -> int:
        """Returns the number of y channels supported by the model"""
        return 1

    # pytype: disable=invalid-annotation
    def _func(
        self,
        x: Array[("num_samples",), np.float64],
        a: ModelParameter(fixed_to=1, scale_func=_power_x_scale),
        x0: ModelParameter(fixed_to=0, scale_func=lambda x_scale, y_scale, _: x_scale),
        y0: ModelParameter(scale_func=lambda x_scale, y_scale, _: y_scale),
        n: ModelParameter(),
    ) -> Array[("num_samples",), np.float64]:
        """Evaluates the model at a given set of x-axis points and with a given set of
        parameter values and returns the result."""
        assert all(x - x0 >= 0), "`x - x0` must be > 0"
        return a * np.float_power(x - x0, n) + y0

    # pytype: enable=invalid-annotation

    def estimate_parameters(
        self,
        x: Array[("num_samples",), np.float64],
        y: Array[("num_samples",), np.float64],
        model_parameters: Dict[str, ModelParameter],
    ):
        """Set heuristic values for model parameters.

        Typically called during `Fitter.fit`. This method may make use of information
        supplied by the user for some parameters (via the `fixed_to` or
        `user_estimate` attributes) to find initial guesses for other parameters.

        The datasets must be sorted in order of increasing x-axis values and must not
        contain any infinite or nan values. If all parameters of the model allow
        rescaling, then `x`, `y` and `model_parameters` will contain rescaled values.

        :param x: x-axis data, rescaled if allowed.
        :param y: y-axis data, rescaled if allowed.
        :param model_parameters: dictionary mapping model parameter names to their
            metadata, rescaled if allowed.
        """
        # Ensure that y is a 1D array
        y = np.squeeze(y)

        unknowns = {
            param
            for param, param_data in model_parameters.items()
            if param_data.fixed_to is None
        }

        model_parameters["x0"].heuristic = 0.0
        model_parameters["y0"].heuristic = 0.0
        model_parameters["a"].heuristic = 1.0
        model_parameters["n"].heuristic = 1.0

        x0 = model_parameters["x0"].get_initial_value()
        y0 = model_parameters["y0"].get_initial_value()
        a = model_parameters["a"].get_initial_value()
        n = model_parameters["n"].get_initial_value()

        if len(unknowns) == 0 or "x0" in unknowns or len(unknowns) > 2:
            # No heuristic needed / we don't have a good heuristic for this case
            # Fit may well fail if we hit this :(
            pass

        elif unknowns == {"a"}:
            raise ValueError
            x = x - x0
            y = y - y0
            a = y / np.float_power(x, n)
            a = self.param_min_sqrs(x, y, model_parameters, "a", a)[0]
            model_parameters["a"].heuristic = a

        elif unknowns == {"n"}:
            n = self.optimal_n(x, y, model_parameters)[0]
            model_parameters["n"].heuristic = n

        elif unknowns == {"y0"}:
            y0 = y - a * np.float_power(x - x0, n)
            y0 = self.param_min_sqrs(x, y, model_parameters, "y0", y0)[0]
            model_parameters["y0"].heuristic = y0

        elif "a" in unknowns:
            pass  # don't have a great heuristic for these cases

        else:
            assert unknowns == {"y0", "n"}, unknowns

            # Datasets normally taken such that they contain a value of y close to y0
            y0_guesses = np.array(
                [
                    0,
                    np.min(y),
                    np.max(y),
                    np.mean(y),
                ]
            )
            y0_bounds = [
                model_parameters["y0"].lower_bound,
                model_parameters["y0"].upper_bound,
            ]
            y0_bounds = [bound for bound in y0_bounds if np.isfinite(bound)]
            y0_guesses = np.append(y0_guesses, y0_bounds)

            ns = np.zeros_like(y0_guesses)
            costs = np.zeros_like(y0_guesses)
            for idx, y0 in np.ndenumerate(y0_guesses):
                model_parameters["y0"].heuristic = y0
                ns[idx], costs[idx] = self.optimal_n(x, y, model_parameters)

            model_parameters["n"].heuristic = float(ns[np.argmin(costs)])
            model_parameters["y0"].heuristic = float(y0_guesses[np.argmin(costs)])

    def optimal_n(self, x, y, model_parameters):
        """Find the optimal (in the least-squared residuals sense) value of `n`
        based on our current best guesses for the other parameters.

        We use the fact that if `y = x^n` then `n = log(y) / log(x)`. This gives
        us an estimate for `n` at each value of x. We choose the one which results
        in lowest sum of squares residuals.
        """
        if model_parameters["n"].has_user_initial_value():
            return model_parameters["n"].get_initial_value(), 0

        x0 = model_parameters["x0"].get_initial_value()
        y0 = model_parameters["y0"].get_initial_value()
        a = model_parameters["a"].get_initial_value()

        x_pr = x - x0
        y_pr = y - y0
        y_pr = y_pr / a

        # Avoid divide by zero errors
        valid = np.argwhere(np.logical_and(y_pr != 0, x_pr != 1))

        if len(valid) == 0:
            return 0, np.inf

        n = np.log(np.abs(y_pr[valid])) / np.log(np.abs(x_pr[valid]))
        n = n.squeeze()

        # Don't look for silly values of n
        n_min = max(-10, model_parameters["n"].lower_bound)
        n_max = min(10, model_parameters["n"].upper_bound)

        n = n[np.argwhere(np.logical_and(n >= n_min, n <= n_max))]

        if len(n) == 0:
            return 1, np.inf

        return self.param_min_sqrs(x, y, model_parameters, "n", n)


def poly_fit_parameter(n):
    def scale_func(x_scale, y_scale, _):
        return y_scale / np.power(x_scale, n)

    return ModelParameter(
        fixed_to=None,
        scale_func=scale_func,
    )


def _poly_x_scale(x_scale: float, y_scale: float, model: Model) -> Optional[float]:
    # If a non-zero value has been set for `x0` we can't rescale the dataset
    if model.parameters["x0"].fixed_to == 0.0:
        return 1
    return None


def _generate_poly_parameters(poly_degree):
    params = {f"a_{n}": poly_fit_parameter(n) for n in range(poly_degree + 1)}
    params.update({"x0": ModelParameter(fixed_to=0, scale_func=_poly_x_scale)})
    return params


class Polynomial(Model):
    """A polynomial fit model.

    Fits the function:
    y = sum(a_n*(x-x0)^n) for n ={0...poly_degree}

    Fit parameters (all floated by default unless stated otherwise):
      - a_0 ... a_{poly_degree}: polynomial coefficients.
      - x0: x-axis offset (fixed to 0 by default). Floating x0 as well as polynomial
          coefficients results in an under-defined problem.

    Derived parameters:
        None
    """

    def __init__(self, poly_degree=10):
        """Init

        :param poly_degree: The degree of the polynomial that we're fitting
           defaults to 10.
        """
        self.poly_degree = poly_degree
        super().__init__(parameters=_generate_poly_parameters(poly_degree))

    def get_num_y_channels(self) -> int:
        """Returns the number of y channels supported by the model"""
        return 1

    def func(
        self, x: Array[("num_samples",), np.float64], params: Dict[str, float]
    ) -> Array[("num_samples",), np.float64]:
        """Evaluates the model at a given set of x-axis points and with a given
        parameter set and returns the result."""
        x0 = params["x0"]
        p = np.array(
            [params[f"a_{n}"] for n in range(self.poly_degree, -1, -1)],
            dtype=np.float64,
        )

        y = np.polyval(p, x - x0)
        return y

    def estimate_parameters(
        self,
        x: Array[("num_samples",), np.float64],
        y: Array[("num_samples",), np.float64],
        model_parameters: Dict[str, ModelParameter],
    ):
        """Set heuristic values for model parameters.

        Typically called during `Fitter.fit`. This method may make use of information
        supplied by the user for some parameters (via the `fixed_to` or
        `user_estimate` attributes) to find initial guesses for other parameters.

        The datasets must be sorted in order of increasing x-axis values and must not
        contain any infinite or nan values. If all parameters of the model allow
        rescaling, then `x`, `y` and `model_parameters` will contain rescaled values.

        :param x: x-axis data, rescaled if allowed.
        :param y: y-axis data, rescaled if allowed.
        :param model_parameters: dictionary mapping model parameter names to their
            metadata, rescaled if allowed.
        """
        # Ensure that y is a 1D array
        y = np.squeeze(y)

        model_parameters["x0"].heuristic = 0.0
        x0 = model_parameters["x0"].get_initial_value()

        free = [
            n
            for n in range(self.poly_degree + 1)
            if model_parameters[f"a_{n}"].fixed_to != 0.0
        ]
        if len(free) == 0:
            return

        deg = max(free)

        p = np.polyfit(x - x0, y, deg)
        for idx in range(deg + 1):
            model_parameters[f"a_{idx}"].heuristic = p[deg - idx]


class Line(MappedModel):
    """Straight line fit according to:
    `y = a * x + y0`

    Fit parameters (all floated by default unless stated otherwise):
      - y0: y-axis intercept
      - a: slope

    Derived parameters:
        None
    """

    def __init__(self):
        super().__init__(
            Polynomial(1),
            {"a": "a_1", "y0": "a_0"},
            {"x0": 0},
        )


class Parabola(MappedModel):
    """Parabola fit according to:
    `y = k * (x - x0)^2 + y0`

    Fit parameters (all floated by default unless stated otherwise):
      - x0: x-axis offset
      - y0: y-axis intercept
      - k: curvature

    Derived parameters:
        None
    """

    def __init__(self):
        inner = Polynomial(2)
        inner.parameters["x0"].fixed_to = None
        inner.parameters["a_2"].fixed_to = None
        super().__init__(
            inner,
            {"k": "a_2", "y0": "a_0", "x0": "x0"},
            {"a_1": 0},
        )

    def _inner_estimate_parameters(
        self,
        x: Array[("num_samples",), np.float64],
        y: Array[("num_samples",), np.float64],
        inner_parameters: Dict[str, ModelParameter],
    ):
        """
        If `x0` is floated, we map the Polynomial `a_1` coefficient onto a value for
        `x0` according to:
        ```
        y = a_0 + a_2 * x^2
        x -> x - x0: y = a_0 + a_2 * (x - x0)^2
        y = a_0 + a_2*x^2 + a_2 * x0^2 + 2*a_2*x*x0
        y = (a_0 + a_2 * x0^2) + 2*a_2*x*x0 + a_2*x^2

        a_0 -> a_0 + a_2 * x0^2
        a_1 -> 2*a_2*x0 => x0 = a_1/(2*a_2)
        a_2 -> a_2
        ```
        """
        super()._inner_estimate_parameters(x, y, inner_parameters)

        a_0 = inner_parameters["a_0"].get_initial_value()
        a_1 = inner_parameters["a_1"].heuristic
        a_2 = inner_parameters["a_2"].get_initial_value()

        x0 = -a_1 / (2 * a_2)
        inner_parameters["x0"].heuristic = x0
        inner_parameters["a_0"].heuristic = a_0 - a_2 * x0**2



import numpy as np
from typing import List, Union


class _SubscriptableNumpyArray(type):
    def __getitem__(cls, _):
        return np.ndarray


class Array(np.ndarray, metaclass=_SubscriptableNumpyArray):
    """
    Subclass of numpy's NDArray used purely for type annotation convenience
    Type annotations can have arbitrary subscripts, e.g. Array[(4,), "float64"].
    """


class _SubscriptableArrayLikeUnion(type):
    def __getitem__(cls, x):
        return Union[List, Array[x]]


class ArrayLike(metaclass=_SubscriptableArrayLikeUnion):
    """
    Used for types that can be a numpy array or a list that's then converted to an array
    """



from typing import Dict, Tuple, TYPE_CHECKING
import numpy as np

from .. import Model, ModelParameter
from ..utils import Array

if TYPE_CHECKING:
    num_samples = float


class Exponential(Model):
    """Exponential function according to:
    y(x < x_dead) = y0
    y(x >= x_dead) = y0 + (y_inf-y0)*(1-exp(-(x-x_dead)/tau))

    Fit parameters (all floated by default unless stated otherwise):
      - x_dead: x-axis "dead time" (fixed to 0 by default)
      - y0: initial (x = x_dead) y-axis offset
      - y_inf: y-axis asymptote (i.e. y(x - x_0 >> tau) => y_inf)
      - tau: decay constant

    Derived parameters:
      - x_1_e: x-axis value for 1/e decay including dead time (`x_1_e = x0 + tau`)
    """

    def get_num_y_channels(self) -> int:
        """Returns the number of y channels supported by the model"""
        return 1

    # pytype: disable=invalid-annotation
    def _func(
        self,
        x: Array[("num_samples",), np.float64],
        x_dead: ModelParameter(
            lower_bound=0, fixed_to=0, scale_func=lambda x_scale, y_scale, _: x_scale
        ),
        y0: ModelParameter(scale_func=lambda x_scale, y_scale, _: y_scale),
        y_inf: ModelParameter(scale_func=lambda x_scale, y_scale, _: y_scale),
        tau: ModelParameter(
            lower_bound=0, scale_func=lambda x_scale, y_scale, _: x_scale
        ),
    ) -> Array[("num_samples",), np.float64]:
        y = y0 + (y_inf - y0) * (1 - np.exp(-(x - x_dead) / tau))
        y = np.where(x >= x_dead, y, y0)
        return y

    # pytype: enable=invalid-annotation

    def estimate_parameters(
        self,
        x: Array[("num_samples",), np.float64],
        y: Array[("num_samples",), np.float64],
        model_parameters: Dict[str, ModelParameter],
    ):
        """Set heuristic values for model parameters.

        Typically called during `Fitter.fit`. This method may make use of information
        supplied by the user for some parameters (via the `fixed_to` or
        `user_estimate` attributes) to find initial guesses for other parameters.

        The datasets must be sorted in order of increasing x-axis values and must not
        contain any infinite or nan values. If all parameters of the model allow
        rescaling, then `x`, `y` and `model_parameters` will contain rescaled values.

        :param x: x-axis data, rescaled if allowed.
        :param y: y-axis data, rescaled if allowed.
        :param model_parameters: dictionary mapping model parameter names to their
            metadata, rescaled if allowed.
        """
        # Ensure that y is a 1D array
        y = np.squeeze(y)

        # Exponentials are generally pretty easy to fit so we keep the estimator simple
        model_parameters["x_dead"].heuristic = 0
        model_parameters["y0"].heuristic = y[0]
        model_parameters["y_inf"].heuristic = y[-1]
        model_parameters["tau"].heuristic = x.ptp()

    def calculate_derived_params(
        self,
        x: Array[("num_samples",), np.float64],
        y: Array[("num_samples",), np.float64],
        fitted_params: Dict[str, float],
        fit_uncertainties: Dict[str, float],
    ) -> Tuple[Dict[str, float], Dict[str, float]]:
        """
        Returns dictionaries of values and uncertainties for the derived model
        parameters (parameters which are calculated from the fit results rather than
        being directly part of the fit) based on values of the fitted parameters and
        their uncertainties.

        :param x: x-axis data
        :param y: y-axis data
        :param: fitted_params: dictionary mapping model parameter names to their
            fitted values.
        :param fit_uncertainties: dictionary mapping model parameter names to
            their fit uncertainties.
        :returns: tuple of dictionaries mapping derived parameter names to their
            values and uncertainties.
        """
        derived_params = {"x_1_e": fitted_params["x_dead"] + fitted_params["tau"]}
        derived_uncertainties = {
            "x_1_e": np.sqrt(
                fit_uncertainties["x_dead"] ** 2 + fit_uncertainties["tau"] ** 2
            )
        }
        return derived_params, derived_uncertainties



import dataclasses
import numpy as np
from scipy import signal
from typing import Dict, Tuple, Type, TYPE_CHECKING, TypeVar
from ..common import Model, ModelParameter
from ..utils import Array


if TYPE_CHECKING:
    num_samples = float
    num_spectrum_samples = float
    num_y_channels = float

TModel = TypeVar("TModel", bound=Type[Model])


@dataclasses.dataclass
class PeriodicModelParameter(ModelParameter):
    period: float = 1
    offset: float = 0
    lower_bound: float = dataclasses.field(init=False)
    upper_bound: float = dataclasses.field(init=False)

    def __post_init__(self):
        self.lower_bound = 1.5 * self.offset
        self.upper_bound = 1.5 * (self.offset + self.period)

    upper_bound = 1.5 * np.pi

    def clip(self, value: float):
        """Clip value to lie between lower and upper bounds."""
        value = value - self.offset
        return (value % self.period) + self.offset


def rescale_model_x(model_class: TModel, x_scale: float) -> TModel:
    """Rescales the x-axis for a model class.

    This is commonly used to convert models between linear and angular units.

    :param model_class: model class to rescale
    :param x_scale: multiplicative x-axis scale factor. To convert a model that takes
      x in angular units and convert to one that takes x in linear units use
      `x_scale = 2 * np.pi`
    """

    class ScaledModel(model_class):
        def __init__(self, *args, **kwargs):
            super().__init__(*args, **kwargs)
            self.__x_scale = x_scale
            self.__rescale = True

        def func(
            self, x: Array[("num_samples",), np.float64], param_values: Dict[str, float]
        ) -> Array[("num_samples", "num_y_channels"), np.float64]:
            x = (x * self.__x_scale) if self.__rescale else x
            return super().func(x, param_values)

        def estimate_parameters(
            self,
            x: Array[("num_samples",), np.float64],
            y: Array[("num_samples", "num_y_channels"), np.float64],
            model_parameters: Dict[str, ModelParameter],
        ):
            # avoid double rescaling if estimate_parameters calls self.func internally
            self.__rescale = False
            super().estimate_parameters(x * self.__x_scale, y, model_parameters)
            self.__rescale = True

    ScaledModel.__name__ = model_class.__name__
    ScaledModel.__doc__ = model_class.__doc__

    return ScaledModel


def get_spectrum(
    x: Array[("num_samples",), np.float64],
    y: Array[("num_samples",), np.float64],
    density_units: bool = True,
    trim_dc: bool = False,
) -> Tuple[
    Array[("num_spectrum_samples",), np.float64],
    Array[("num_spectrum_samples",), np.float64],
]:
    """Returns the frequency spectrum (Fourier transform) of a dataset.

    :param x: 1D ndarray of shape (num_samples,) containing x-axis data
    :param y: 1D ndarray of shape (num_samples,) containing y-axis data
    :param density_units: if `False` we apply normalization for narrow-band signals. If
        `True` we normalize for continuous distributions.
    :param trim_dc: if `True` we do not return the DC component.
    """
    if x.ndim != 1:
        raise ValueError("x-axis data must be a 1D array.")
    if y.ndim != 1:
        raise ValueError("y-axis data must be a 1D array.")

    dx = x.ptp() / x.size
    n = x.size
    omega = np.fft.fftfreq(n, dx) * (2 * np.pi)
    y_f = np.fft.fft(y, norm="ortho") / np.sqrt(n)

    y_f = y_f[: int(n / 2)]
    omega = omega[: int(n / 2)]

    if density_units:
        d_omega = 2 * np.pi / (n * dx)
        y_f /= d_omega

    if trim_dc:
        omega = omega[1:]
        y_f = y_f[1:]

    return omega, y_f


def get_pgram(
    x: Array[("num_samples",), np.float64],
    y: Array[("num_samples",), np.float64],
    density_units: bool = True,
) -> Tuple[
    Array[("num_spectrum_samples",), np.float64],
    Array[("num_spectrum_samples",), np.float64],
]:
    """Returns a periodogram for a dataset, converted into amplitude units.

    Based on the Lombe-Scargle periodogram (essentially least-squares fitting of
    sinusoids at different frequencies).

    :param x: x-axis data
    :param y: y-axis data. For models with multiple y channels, this should contain
        data from a single channel only.
    :param density_units: if `False` (default) we apply normalization for narrow-band
        signals. If `True` we normalize for continuous distributions.
    :returns: tuple with the frequency axis (angular units) and the periodogram
    """
    if y.ndim != 1 and y.shape[1] > 1:
        raise ValueError(
            f"{y.shape[1]} y channels were provided to a method which takes 1"
        )

    dx = np.min(np.diff(x))
    duration = x.ptp()
    n = int(duration / dx)
    d_omega = 2 * np.pi / (n * dx)

    # Nyquist limit does not apply to irregularly spaced data
    # We'll use it as a starting point anyway...
    f_nyquist = 0.5 / dx

    omega_list = 2 * np.pi * np.linspace(d_omega, f_nyquist, n)
    pgram = signal.lombscargle(x, y, omega_list, precenter=True)
    pgram = np.sqrt(np.abs(pgram) * 4 / len(y))

    if density_units:
        pgram /= d_omega

    return omega_list, pgram



from typing import Dict, TYPE_CHECKING
import numpy as np

from .rectangle import Rectangle
from .triangle import Triangle
from .utils import get_spectrum
from .. import NormalFitter, Model, ModelParameter
from ..utils import Array

if TYPE_CHECKING:
    num_samples = float


class Sinc(Model):
    """Sinc function according to:
    y = a * sin(w * (x - x0)) / (w * (x - x0)) + y0

    Fit parameters (all floated by default unless stated otherwise):
      - x0: x-axis offset
      - y0: y-axis offset
      - a: amplitude
      - w: x scale factor

    Derived parameters:
      None
    """

    def get_num_y_channels(self) -> int:
        """Returns the number of y channels supported by the model"""
        return 1

    # pytype: disable=invalid-annotation
    def _func(
        self,
        x: Array[("num_samples",), np.float64],
        x0: ModelParameter(scale_func=lambda x_scale, y_scale, _: x_scale),
        y0: ModelParameter(scale_func=lambda x_scale, y_scale, _: y_scale),
        a: ModelParameter(scale_func=lambda x_scale, y_scale, _: y_scale),
        w: ModelParameter(
            lower_bound=0, scale_func=lambda x_scale, y_scale, _: 1 / x_scale
        ),
    ) -> Array[("num_samples",), np.float64]:
        x = w * (x - x0) / np.pi  # np.sinc(x) = sin(pi*x) / (pi*x)
        y = a * np.sinc(x) + y0
        return y

    # pytype: enable=invalid-annotation
    def estimate_parameters(
        self,
        x: Array[("num_samples",), np.float64],
        y: Array[("num_samples",), np.float64],
        model_parameters: Dict[str, ModelParameter],
    ):
        """Sets initial values for model parameters based on heuristics. Typically
        called during `Fitter.fit`.

        Heuristic results should be stored in :param model_parameters: using the
        `ModelParameter`'s `heuristic` attribute. This ensures that all information
        passed in by the user (fixed values, initial values, bounds) is used correctly.

        The dataset must be sorted in order of increasing x-axis values and must not
        contain any infinite or nan values.

        :param x: x-axis data
        :param y: y-axis data
        :param model_parameters: dictionary mapping model parameter names to their
            metadata.
        """
        # Ensure that y is a 1D array
        y = np.squeeze(y)

        model_parameters["y0"].heuristic = np.mean([y[0], y[-1]])
        y0 = model_parameters["y0"].get_initial_value()

        omega, spectrum = get_spectrum(x, y, trim_dc=True)
        abs_spectrum = np.abs(spectrum)

        # Fourier transform of a sinc is a rectangle
        rect = Rectangle()
        rect.parameters["y0"].fixed_to = 0
        rect.parameters["x_l"].fixed_to = 0
        rect.parameters["a"].heuristic = max(abs_spectrum)

        fit = NormalFitter(omega, abs_spectrum, model=rect)

        model_parameters["w"].heuristic = fit.values["x_r"]
        w = model_parameters["w"].get_initial_value()

        sgn = 1 if y[np.argmax(np.abs(y - y0))] > y0 else -1
        model_parameters["a"].heuristic = 2 * w * fit.values["a"] * sgn

        x0 = self.find_x_offset_sym_peak(
            x=x,
            y=y,
            parameters=model_parameters,
            omega=omega,
            spectrum=spectrum,
            omega_cut_off=w,
        )

        model_parameters["x0"].heuristic = x0


class Sinc2(Model):
    """Sinc-squared function according to:
    y = a * (sin(w * (x - x0)) / (w * (x - x0)))^2 + y0

    Fit parameters (all floated by default unless stated otherwise):
      - x0: x-axis offset
      - y0: y-axis offset
      - a: amplitude
      - w: x scale factor

    Derived parameters:
      None
    """

    def get_num_y_channels(self) -> int:
        """Returns the number of y channels supported by the model"""
        return 1

    # pytype: disable=invalid-annotation
    def _func(
        self,
        x: Array[("num_samples",), np.float64],
        x0: ModelParameter(scale_func=lambda x_scale, y_scale, _: x_scale),
        y0: ModelParameter(scale_func=lambda x_scale, y_scale, _: y_scale),
        a: ModelParameter(scale_func=lambda x_scale, y_scale, _: y_scale),
        w: ModelParameter(
            lower_bound=0, scale_func=lambda x_scale, y_scale, _: 1 / x_scale
        ),
    ) -> Array[("num_samples",), np.float64]:
        x = w * (x - x0) / np.pi  # np.sinc(x) = sin(pi*x) / (pi*x)
        y = a * np.power(np.sinc(x), 2) + y0
        return y

    # pytype: enable=invalid-annotation

    def estimate_parameters(
        self,
        x: Array[("num_samples",), np.float64],
        y: Array[("num_samples",), np.float64],
        model_parameters: Dict[str, ModelParameter],
    ):
        """Sets initial values for model parameters based on heuristics. Typically
        called during `Fitter.fit`.

        Heuristic results should stored in :param model_parameters: using the
        `ModelParameter`'s `initialise` method. This ensures that all information passed
        in by the user (fixed values, initial values, bounds) is used correctly.

        The dataset must be sorted in order of increasing x-axis values and must not
        contain any infinite or nan values.

        :param x: x-axis data
        :param y: y-axis data
        :param model_parameters: dictionary mapping model parameter names to their
            metadata.
        """
        # Ensure that y is a 1D array
        y = np.squeeze(y)

        y0 = model_parameters["y0"].heuristic = np.mean([y[0], y[-1]])

        omega, spectrum = get_spectrum(x, y, trim_dc=True)
        abs_spectrum = np.abs(spectrum)

        # Fourier transform of a sinc^2 is a triangle function
        tri = Triangle()
        tri.parameters["x0"].fixed_to = 0
        tri.parameters["y0"].heuristic = max(abs_spectrum)
        tri.parameters["sym"].fixed_to = 0
        tri.parameters["y_min"].fixed_to = 0

        fit = NormalFitter(omega, abs_spectrum, model=tri)

        intercept = fit.values["y0"] / -fit.values["k"]
        sgn = 1 if y[np.argmax(np.abs(y - y0))] > y0 else -1

        model_parameters["w"].heuristic = 0.5 * intercept
        model_parameters["a"].heuristic = fit.values["y0"] * sgn * intercept

        x0 = self.find_x_offset_sym_peak(
            x=x,
            y=y,
            parameters=model_parameters,
            omega=omega,
            spectrum=spectrum,
            omega_cut_off=intercept,
        )

        model_parameters["x0"].heuristic = x0



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['ionics_fits', 'ionics_fits.models']

package_data = \
{'': ['*']}

install_requires = \
['numpy>=1.21.2,<2.0.0',
 'pyqt5-qt5>=5.15.2,<6.0.0',
 'pyqt5>=5.15.9,<6.0.0',
 'scipy>=1.7.1,<2.0.0',
 'statsmodels>=0.13.2,<0.14.0']

setup_kwargs = {
    'name': 'ionics-fits',
    'version': '1.0.10',
    'description': 'Lightweight Python data fitting library with an emphasis on AMO/Quantum Information',
    'long_description': 'Lightweight Python library for data fitting with an emphasis on AMO (Atomic Molecular\nand Optical physics) and Quantum Information.\n\n`fits` was inspired by the [Oxford Ion Trap Group fitting library](https://github.com/OxfordIonTrapGroup/oitg/tree/master/oitg/fitting) originally\nauthored by @tballance. It is still in the alpha phase and is likely to be renamed\n(although `fits` is still available on pypi)...please feel free to help bikeshed names.\n\n# Getting started\n\n## Installation\n\nInstall from `pypi` with `pip install ionics_fits` or add to your poetry project with\n`poetry add fits`.\n\n## Example Usage\n\nBasic usage\n```python\nimport numpy as np\nimport ionics_fits as fits\n\na = 3.2\ny0 = -9\n\nx = np.linspace(-10, 10)\ny = a*x + y0\n\nfit = fits.NormalFitter(x, y, model=fits.models.Line())\nprint(f"Fitted: y = {fit.values[\'a\']:.3f} * x + {fit.values[\'y0\']:.3f}")\n```\n\nThe fit can be configured in various ways by modifying the model\'s `parameters`\ndictionary (see the `fits.common.ModelParameter` class for more information). This\nallows one to:\n- change the bounds for parameters\n- change which parameters are fixed to a constant value / floated\n- supply initial values for parameters instead of relying on the heuristics\n\nExample usage:\n```python\nimport numpy as np\nfrom matplotlib import pyplot as plt\nimport ionics_fits as fits\n\n# Example problem: fit the amplitude and phase of a sinusoid whose frequency is known\n# exactly\n\nomega = 2 * np.pi  # we know the frequency\nmodel = fits.models.Sinusoid()\nmodel.parameters["omega"].fixed_to = omega\n\n# generate synthetic data to fit\nparams = {\n    "a": np.random.uniform(low=1, high=5),\n    "omega": omega,\n    "phi": np.random.uniform(-1, 1) * 2 * np.pi,\n    "y0": 0,\n    "x0": 0,\n    "tau": np.inf,\n}\nx = np.linspace(-3, 3, 100)\ny = model.func(x, params)\n\nfit = fits.NormalFitter(x, y, model=model)\nprint(f"Amplitude: dataset = {params[\'a\']:.3f}, fit = {fit.values[\'a\']:.3f}")\nprint(f"Phase: dataset = {params[\'phi\']:.3f}, fit = {fit.values[\'phi\']:.3f}")\n\nplt.plot(x, y, label="data")\nplt.plot(*fit.evaluate(True), \'-.o\', label="fit")\nplt.grid()\nplt.legend()\nplt.show()\n```\n\n# Developing\n\nBefore committing:\n- Update formatting: `poe fmt`\n- Lints: `poe flake`\n- Run test suite: `poe test`\n- Optionally: [fuzz](#Fuzzing) any new models\n\n# Design Philosophy\n\n## Good Heuristics\n\nLife is too short for failed fits. We can\'t guarantee to fit every dataset without any\nhelp from the user (e.g. specifying initial parameter values) no matter how noisy or\nincomplete it is...but we do our best!\n\nEvery fit model has a "parameter estimator" which uses heuristics to find good estimates\nof the values of the model\'s free parameters. We expect these heuristics to be good\nenough to allow the optimizer to fit any "reasonable" dataset. Fit failures are viewed\nas a bug and we encourage our users to file issues where they find them (please post an\nexample dataset in the issue).\n\nCurrently this project is a MVP and many of the heuristics need some work. Expect there\nto be cases where we could easily do better. Please report them where you find them!\n\n## Validation\n\nIt\'s not enough to just fit the data, we want to know if we can trust the fit results\nbefore acting on them.  There are two distinct aspects to the validation problem: did\nthe fit find the model parameters which best match the data (as opposed to getting stuck\nin a local minimum in parameter space far from the global optimum)? and, are the fitted\nparameter values consistent with our prior knowledge of the system (e.g. we know that a\nfringe contrast must lie within certain bounds).\n\nFirst, any prior knowledge about the system should be incorporated by specifying fixed\nparameter values and parameter bounds. After that, the fit is validated. At present,\nvalidation is done using the Chi-squared as a test for goodness of fit. It is likely\nthat additional validation tests will be added as the package grows.\n\n## General purpose\n\nThis library is designed to be general purpose; rather than tackling specific problems\nwe try to target sets of problems -- we want to fit sinusoids not *your* sinusoid. This\nis reflected, for example, in the choices of parametrisation, which are intended to be\nextremely flexible, and the effort put into heuristics. If you find you can\'t easily fit\nyour sinusoid with the standard model/heuristics it\'s probably a bug in the model design\nso please open an issue.\n\nWe encourage contributions of new fit models, but please consider generality before\nsubmission. If you want to solve a specific problem in front of you, that\'s fine but\nprobably better suited to your own codebase.\n\n## Extensibility\n\nThe library is designed to be extensible and ergonomic to user. Want to use different\nstatistics? Easy, just provide a new class that inherits from `FitBase`. Want to do some\ncustom post-fit processing? Override the `calculate_derived_parameters` method. Want to\ntweak the parameter estimator for a model? Create a new model class that inherits from\nthe original model and modify away. If you\'re struggling to do what you want, it\'s\nprobably a bug in the library so please report it.\n\n`ionics_fits` provides a number of tools to make it easier to extend models. See, for\nexample [`models.utils`](../master/ionics_fits/models/utils.py) and [`models.containers`](../master/ionics_fits/models/containers.py). Suppose you want to...\n## Rescaling models\n\n...fit some frequency-domain Rabi oscillation data. However, the model works in angular\nunits, but your tooling needs linear units. No problem! Simply use the `rescale_model_x`\ntool.\n\n```python\ndetuning_model = fits.models.utils.rescale_model_x(fits.models.RabiFlopFreq, 2 * np.pi)\n```\n\nOr, suppose you actually want to scan the magnetic field and find the field offset which\nputs the transition at a particular frequency?\n```python\nclass _RabiBField(fits.models.RabiFlopFreq):\n    def __init__(self, dfdB, B_0, f_0, start_excited):\n        super().__init__(start_excited=start_excited)\n        self.dfdB = dfdB\n        self.B_0 = B_0\n        self.f_0 = f_0\n\n    def calculate_derived_params(self, x, y, fitted_params, fit_uncertainties):\n        derived_params, derived_uncertainties = super().calculate_derived_params(\n            x, y, fitted_params, fit_uncertainties\n        )\n\n        df = derived_params["f_0"] - self.f_0\n        dB = df / self.dfdB\n        B_0 = dB + self.B_0\n\n        derived_params["B_0"] = B_0\n        derived_uncertainties["B_0"] = derived_uncertainties["f_0"] * self.dfdB\n\n        return derived_params, derived_uncertainties\n\nRabiBField = fits.models.utils.rescale_model_x(_RabiBField, 2 * np.pi * dfdB)\n```\n\n## Containers\n\n...fit multiple independent models simultaneously and do some post-processing on the\nresults. Use an `AggregateModel`.\n\n```python\nclass LineAndTriange(fits.models.AggregateModel):\n  def __init__(self):\n    line = fits.models.Line()\n    triangle = fits.models.Triangle()\n    super().__init__(models=[("line", line), "triangle", triangle])\n\n  def calculate_derived_params(\n      self,\n      x: Array[("num_samples",), np.float64],\n      y: Array[("num_samples",), np.float64],\n      fitted_params: Dict[str, float],\n      fit_uncertainties: Dict[str, float],\n  ) -> Tuple[Dict[str, float], Dict[str, float]]:\n      derived_params, derived_uncertainties = super().calculate_derived_params()\n      # derive new results from the two fits\n      return derived_params, derived_uncertainties\n```\n\n...use the single-qubit Rabi flop model to fit simultaneous Rabi flopping on multiple\nqubits at once with some parameters shared and some independent.\n\n```python\nclass MultiRabiFreq(fits.models.RepeatedModel):\n    def __init__(self, n_qubits):\n        super().__init__(\n            inner=fits.models.RabiFlopFreq(start_excited=True),\n            common_params=[\n                "P_readout_e",\n                "P_readout_g",\n                "t_pulse",\n                "omega",\n                "tau",\n                "t_dead",\n            ],\n            num_repetitions=n_qubits,\n        )\n\n```\n## And more!\n\nAt present the library is still an MVP. Further work will be driven by use cases, so\nplease open an issue if you find you can\'t easily extend the library in the way you\nwant.\n\n# Ontology\n\nThere are two main kinds of object in the library: `fit`s and `model`s. Models are\ngeneral-purpose functions to be fitted, such as sinusoids or Lorentzians, but are\nagnostic about the statistics. Fits do the fitting (maximum likelihood parameter\nestimation) and validation based on some underlying statistics (normal, binomial, etc). \n\n# Testing methodology\n\nThis package uses both `unit test`s and `fuzzing`.\n\n## Unit Tests\n\n- run using `poe test`\n- to run a subset of tests use the `-k` flag e.g. `poe test -k "rabi"` to run only tests\n  with the word `rabi` in their name. For more information about configuring pytest see\n  the [documentation](https://docs.pytest.org/en/7.1.x/)\n- all tests must pass before a PR can be merged into master\n- PRs to add new models will only be merged once they have reasonable test coverage\n- unit tests aim to provide good coverage over the space of "reasonable datasets". There\n  will always be corner-cases where the fits fail and that\'s fine; the aim here is to\n  cover the main cases users will hit in the wild\n- when a user hits a case in the wild where the fit fails unexpectedly (i.e. we think\n  the fit code should have handled it), a `regression test` based on the failing\n  dataset should be added\n- unit tests should be deterministic. Synthetic datasets should be included in the test\n  rather than randomly generated at run time. Tip: while writing a test it\'s fine to let\n  the test code generate datasets for you. Once you\'re happy, run the test in verbose\n  mode and copy the dataset from the log output\n\n## Fuzzing\n\n- fuzzing is non-deterministic (random parameter values, randomly generated datasets)\n  exploration of the parameter space.\n- used when developing / debugging fits, but not automatically run by CI\n- run with `poe fuzz` (see `--help` for details)\n- fit failures during fuzzing are not automatically considered a bug; unlike unit tests,\n  fuzzing explores the "unreasonable" part of parameter space as well. Indeed, a large\n  part of the point of fuzzing is to help the developer understand what should be\n  considered "reasonable" (this information should end up in the documentation for the\n  fuzzed model).\n',
    'author': 'hartytp',
    'author_email': 'thomas.peter.harty@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<3.11',
}


setup(**setup_kwargs)



import copy
from typing import Dict, Tuple, TYPE_CHECKING
import numpy as np

from .. import Model, ModelParameter
from ..utils import Array

if TYPE_CHECKING:
    num_samples = float


class Triangle(Model):
    """Triangle function according to:
    y(x>=x0) = k_p*|x-x0| + y0
    y(x<x0) = k_m*|x-x0| + y0
    y = max(y, y_min)
    y = min(y, m_max)
    k_p = (1 + sym) * k
    k_m = (1 - sym) * k

    Fit parameters (all floated by default unless stated otherwise):
      - x0: x-axis offset
      - y0: y-axis offset
      - k: average slope
      - sym: symmetry parameter (fixed to 0 by default)
      - y_min: minimum value of y (bound to -inf by default)
      - y_max: maximum value of y (bound to +inf by default)

    Derived parameters:
      - k_m: slope for x < x0
      - k_p: slope for x >= x0
    """

    def get_num_y_channels(self) -> int:
        """Returns the number of y channels supported by the model"""
        return 1

    # pytype: disable=invalid-annotation
    def _func(
        self,
        x: Array[("num_samples",), np.float64],
        x0: ModelParameter(scale_func=lambda x_scale, y_scale, _: x_scale),
        y0: ModelParameter(scale_func=lambda x_scale, y_scale, _: y_scale),
        k: ModelParameter(scale_func=lambda x_scale, y_scale, _: y_scale / x_scale),
        sym: ModelParameter(lower_bound=-1, upper_bound=1, fixed_to=0),
        y_min: ModelParameter(
            fixed_to=-np.inf, scale_func=lambda x_scale, y_scale, _: y_scale
        ),
        y_max: ModelParameter(
            fixed_to=+np.inf, scale_func=lambda x_scale, y_scale, _: y_scale
        ),
    ) -> Array[("num_samples",), np.float64]:
        k_p = k * (1 + sym)
        k_m = k * (1 - sym)

        y_p = k_p * np.abs(x - x0) + y0
        y_m = k_m * np.abs(x - x0) + y0
        y = np.where(x >= x0, y_p, y_m)
        y = np.where(y > y_min, y, y_min)
        y = np.where(y < y_max, y, y_max)

        return y

    # pytype: enable=invalid-annotation
    def estimate_parameters(
        self,
        x: Array[("num_samples",), np.float64],
        y: Array[("num_samples",), np.float64],
        model_parameters: Dict[str, ModelParameter],
    ):
        """Set heuristic values for model parameters.

        Typically called during `Fitter.fit`. This method may make use of information
        supplied by the user for some parameters (via the `fixed_to` or
        `user_estimate` attributes) to find initial guesses for other parameters.

        The datasets must be sorted in order of increasing x-axis values and must not
        contain any infinite or nan values. If all parameters of the model allow
        rescaling, then `x`, `y` and `model_parameters` will contain rescaled values.

        :param x: x-axis data, rescaled if allowed.
        :param y: y-axis data, rescaled if allowed.
        :param model_parameters: dictionary mapping model parameter names to their
            metadata, rescaled if allowed.
        """
        # Ensure that y is a 1D array
        y = np.squeeze(y)

        # Written to handle the case of data which is only well-modelled by a
        # triangle function near `x0` but saturates further away
        model_parameters["y_max"].heuristic = max(y)
        model_parameters["y_min"].heuristic = min(y)

        min_ind = np.argmin(y)
        max_ind = np.argmax(y)

        y_min = y[min_ind]
        y_max = y[max_ind]
        x_min = x[min_ind]
        x_max = x[max_ind]

        # Case 1: positive slope with peaks left and right of x_min above y_min
        x_l = x[x <= x_min]
        x_r = x[x >= x_min]
        y_l = y[x <= x_min]
        y_r = y[x >= x_min]

        left_peak_ind = np.argmax(y_l)
        right_peak_ind = np.argmax(y_r)

        dx_l = x_min - x_l[left_peak_ind]
        dx_r = x_r[right_peak_ind] - x_min

        k_m = (y_l[left_peak_ind] - y_min) / dx_l if dx_l != 0 else 0
        k_p = (y_r[right_peak_ind] - y_min) / dx_r if dx_r != 0 else 0
        alpha = 0 if k_m == 0 else k_p / k_m

        positive_parameters = copy.deepcopy(model_parameters)
        positive_parameters["x0"].heuristic = x_min
        positive_parameters["y0"].heuristic = y_min

        positive_parameters["k"].heuristic = 0.5 * (k_p + k_m)
        positive_parameters["sym"].heuristic = (alpha - 1) / (1 + alpha)

        positive_parameters = {
            param: param_data.get_initial_value()
            for param, param_data in positive_parameters.items()
        }
        positive_residuals = np.sum(
            np.power(y - self._func(x, **positive_parameters), 2)
        )

        # Case 2: negative slope with peaks left and right of x_max below y_max
        x_l = x[x <= x_max]
        x_r = x[x >= x_max]
        y_l = y[x <= x_max]
        y_r = y[x >= x_max]

        left_peak_ind = np.argmin(y_l)
        right_peak_ind = np.argmin(y_r)

        dx_l = x_max - x_l[left_peak_ind]
        dx_r = x_r[right_peak_ind] - x_max

        k_m = (y_l[left_peak_ind] - y_max) / dx_l if dx_l != 0 else 0
        k_p = (y_r[right_peak_ind] - y_max) / dx_r if dx_r != 0 else 0
        alpha = 0 if k_m == 0 else k_p / k_m

        negative_parameters = copy.deepcopy(model_parameters)
        negative_parameters["x0"].heuristic = x_max
        negative_parameters["y0"].heuristic = y_max
        negative_parameters["k"].heuristic = 0.5 * (k_p + k_m)
        negative_parameters["sym"].heuristic = (alpha - 1) / (1 + alpha)

        negative_parameters = {
            param: param_data.get_initial_value()
            for param, param_data in negative_parameters.items()
        }
        negative_residuals = np.sum(
            np.power(y - self._func(x, **negative_parameters), 2)
        )

        if positive_residuals < negative_residuals:
            best_params = positive_parameters
        else:
            best_params = negative_parameters

        for param, value in best_params.items():
            model_parameters[param].heuristic = value

    def calculate_derived_params(
        self,
        x: Array[("num_samples",), np.float64],
        y: Array[("num_samples",), np.float64],
        fitted_params: Dict[str, float],
        fit_uncertainties: Dict[str, float],
    ) -> Tuple[Dict[str, float], Dict[str, float]]:
        """
        Returns dictionaries of values and uncertainties for the derived model
        parameters (parameters which are calculated from the fit results rather than
        being directly part of the fit) based on values of the fitted parameters and
        their uncertainties.

        :param x: x-axis data
        :param y: y-axis data
        :param: fitted_params: dictionary mapping model parameter names to their
            fitted values.
        :param fit_uncertainties: dictionary mapping model parameter names to
            their fit uncertainties.
        :returns: tuple of dictionaries mapping derived parameter names to their
            values and uncertainties.
        """
        derived_params = {}
        k = fitted_params["k"]
        sym = fitted_params["sym"]

        k_err = fitted_params["k"]
        sym_err = fitted_params["sym"]

        derived_params["k_p"] = (1 + sym) * k
        derived_params["k_m"] = (1 - sym) * k

        derived_uncertainties = {}
        derived_uncertainties["k_p"] = np.sqrt(
            (k_err * (1 + sym)) ** 2 + (k * sym_err) ** 2
        )
        derived_uncertainties["k_m"] = np.sqrt(
            (k_err * (1 - sym)) ** 2 + (k * sym_err) ** 2
        )

        return derived_params, derived_uncertainties
